import pandas as pd
import unittest
import pyterrier as pt
import warnings
from .base import BaseTestCase
from pytest import warns

class TestOperators(BaseTestCase):

    def test_then_dataframe(self):
        #this test is we can have DataFrame >> SOMETHINGELSE

        topicsSource = pd.DataFrame([["1", "AA"]], columns=["qid", "query"])

        def rewrite(topics):
            topics = topics.copy()
            for index, row in topics.iterrows():
                row["query"] = row["query"] + " test"
            return topics
        fn1 = lambda topics : rewrite(topics)
        import pyterrier.transformer as ptt
        
        topics = pd.DataFrame([["1", "A"]], columns=["qid", "query"])
        rtr = pt.Transformer.from_df(topicsSource)(topics)
        self.assertEqual(1, len(rtr))
        self.assertTrue("query" in rtr.columns)
        self.assertTrue("qid" in rtr.columns)
        self.assertEqual(2, len(rtr.columns))  
        self.assertEqual("AA", rtr.iloc[0]["query"])

        sequence1 = pt.Transformer.from_df(topicsSource) >> pt.apply.generic(fn1)
        rtr = sequence1(topics)
        self.assertTrue("query" in rtr.columns)
        self.assertTrue("qid" in rtr.columns)
        self.assertEqual(2, len(rtr.columns))        
        self.assertEqual(1, len(rtr))
        self.assertEqual("AA test", rtr.iloc[0]["query"])

    def test_then(self):
        
        def rewrite(topics):
            topics = topics.copy()
            for index, row in topics.iterrows():
                row["query"] = row["query"] + " test"
            return topics
        fn1 = lambda topics : rewrite(topics)
        fn2 = lambda topics : rewrite(topics)
        import pyterrier.apply_base as ptt
        sequence1 = ptt.ApplyGenericTransformer(fn1) >> ptt.ApplyGenericTransformer(fn2)

        
        for sequence in [sequence1]:
            self.assertTrue(isinstance(sequence, pt.Transformer))
            #check we can access items
            self.assertEqual(2, len(sequence))
            self.assertTrue(sequence[0], pt.Transformer)
            self.assertTrue(sequence[1], pt.Transformer)
            input = pd.DataFrame([["q1", "hello"]], columns=["qid", "query"])
            output = sequence.transform(input)
            self.assertEqual(1, len(output))
            self.assertEqual("q1", output.iloc[0]["qid"])
            self.assertEqual("hello test test", output.iloc[0]["query"])
            # now test transform_iter pathway via __call__
            output = sequence(input.to_dict(orient='records'))
            self.assertIsInstance(output, list)
            output = pd.DataFrame(output)
            self.assertEqual(1, len(output))
            self.assertEqual("q1", output.iloc[0]["qid"])
            self.assertEqual("hello test test", output.iloc[0]["query"])

    def test_compile_fuse_right(self):
        from typing import Optional

        class B(pt.Transformer):
            def transform(self, df):
                pass

        class C(pt.Transformer):
            def transform(self, df):
                pass

        class A(pt.Transformer):
            def transform(self, df):
                pass
            def fuse_right(self, right: pt.Transformer) -> Optional[pt.Transformer]:
                if isinstance(right, B):
                    return C()
                return None

        test1 = A()
        self.assertIsInstance(test1.compile(), A)

        test2 = A() >> A()
        test2_c = test2.compile()
        self.assertEqual(2, len(test2_c))
        self.assertIsInstance(test2_c[0], A)
        self.assertIsInstance(test2_c[1], A)

        test3 = A() >> B()
        test3_c = test3.compile()
        self.assertIsInstance(test3_c, C)

    def test_compile_fuse_left(self):
        from typing import Optional

        class B(pt.Transformer):
            def transform(self, df):
                pass

            def fuse_left(self, left: pt.Transformer) -> Optional[pt.Transformer]:
                if isinstance(left, A):
                    return C()
                return None

        class C(pt.Transformer):
            def transform(self, df):
                pass

        class A(pt.Transformer):
            def transform(self, df):
                pass

        test1 = B()
        self.assertIsInstance(test1.compile(), B)

        test2 = B() >> B()
        test2_c = test2.compile()
        self.assertEqual(2, len(test2_c))
        self.assertIsInstance(test2_c[0], B)
        self.assertIsInstance(test2_c[1], B)

        test3 = A() >> B()
        test3_c = test3.compile()
        self.assertIsInstance(test3_c, C)

    def test_compile(self):
        class PRF(pt.Transformer):
            def __init__(self, k=10):
                self.k = k
            def transform(self, df):
                pass
            def compile(self):
                # TODO: this shouldnt be so hacky
                import pyterrier._ops
                return pyterrier._ops.RankCutoff(self.k) >> self
            def __repr__(self):
                return f'PRF(k={self.k})'
                
        class Retr(pt.Transformer):
            def __init__(self, k=1000):
                self.k = k
            def fuse_rank_cutoff(self, k):
                return Retr(k=k)
            def transform(self, df):
                pass
            def __repr__(self):
                return f'Retr(k={self.k})'

        class Extractor(pt.Transformer):
            def transform(self, df):
                pass
            def fuse_rank_cutoff(self, k):
                # TODO: this shouldnt be so hacky
                import pyterrier._ops
                return pyterrier._ops.RankCutoff(k) >> self
            def __repr__(self):
                return 'Extr'

        # check that the rank cutoff fusion works.
        pipe1 = Retr() % 3
        pipe1c = pipe1.compile()
        self.assertEqual(3, pipe1c.k)
        self.assertIsInstance(pipe1c, Retr)

        # now check that we can propagate k=3 (from PRF.compile()) all the way back to Retr
        slow_pipe = Retr() >> Extractor() >> PRF(k=3) >> Retr()
        fast_pipe = slow_pipe.compile()
        # the resulting pipe should be:
        # Retr(3) >> Extr() >> PRF(k=3) >> Retr(1000)
        self.assertEqual(4, len(fast_pipe))
        self.assertEqual(Retr, type(fast_pipe[0]))
        self.assertEqual(3, fast_pipe[0].k)
        self.assertEqual(Extractor, type(fast_pipe[1]))
        self.assertEqual(PRF, type(fast_pipe[2]))
        self.assertEqual(3, fast_pipe[2].k)
        self.assertEqual(Retr, type(fast_pipe[3]))
        self.assertEqual(1000, fast_pipe[3].k)

    def test_then_multi(self):
        import pyterrier.transformer as ptt
        mock1 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 5]], columns=["qid", "docno", "score"]), uniform=True)
        mock2 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 10]], columns=["qid", "docno", "score"]), uniform=True)
        mock3 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 10]], columns=["qid", "docno", "score"]), uniform=True)
        mock4 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 10]], columns=["qid", "docno", "score"]), uniform=True)
        
        combined12 = mock1 >> mock2
        combined23 = mock2 >> mock3 
        combined123_a = combined12 >> mock3
        combined123_b = mock1 >> mock2 >> mock3
        combined123_c = mock2 >> combined23

        combined123_a_C = combined123_a.compile()
        combined123_b_C = combined123_b.compile()
        combined123_c_C = combined123_c.compile()


        self.assertEqual(2, len(combined12))
        self.assertEqual(2, len(combined23))
        self.assertEqual(2, len(combined12))
        self.assertEqual(2, len(combined23))

        for C in [combined123_a_C, combined123_b_C, combined123_c_C]:
            self.assertEqual(3, len(C))
            self.assertEqual("(UniformTransformer() >> UniformTransformer() >> UniformTransformer())", repr(C))
        
        # finally check recursive application
        C4 = (mock1 >> mock2 >> mock3 >> mock4).compile()
        self.assertEqual("(UniformTransformer() >> UniformTransformer() >> UniformTransformer() >> UniformTransformer())", repr(C4))
        self.assertEqual(4, len(C4))


    def test_mul(self):

        import pyterrier.transformer as ptt
        mock = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 5]], columns=["qid", "docno", "score"]), uniform=True)
        for comb in [mock * 10, 10 * mock]:
            rtr = comb.transform(None)
            self.assertEqual(1, len(rtr))
            self.assertEqual("q1", rtr.iloc[0]["qid"])
            self.assertEqual("doc1", rtr.iloc[0]["docno"])
            self.assertEqual(50, rtr.iloc[0]["score"])

        import pyterrier.transformer as ptt
        from pyterrier.model import add_ranks
        mock = pt.Transformer.from_df(add_ranks(pd.DataFrame([["q1", "doc1", 5], ["q1", "doc2", 10]], columns=["qid", "docno", "score"]), single_query=True), uniform=True)
        rtr = mock.search("bla", qid="q1")
        self.assertEqual(2, len(rtr))
        self.assertEqual("q1", rtr.iloc[0]["qid"])
        self.assertEqual("doc2", rtr.iloc[0]["docno"])
        self.assertEqual(pt.model.FIRST_RANK, rtr.iloc[0]["rank"])

        rtr = (-1 * mock).search("bla", qid="q1")
        self.assertEqual(2, len(rtr))
        self.assertEqual("q1", rtr.iloc[0]["qid"])
        self.assertEqual("doc1", rtr.iloc[0]["docno"])
        self.assertEqual(pt.model.FIRST_RANK, rtr.iloc[0]["rank"])
    
    def test_plus(self):
        import pyterrier.transformer as ptt
        mock1 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 5]], columns=["qid", "docno", "score"]), uniform=True)
        mock2 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 10]], columns=["qid", "docno", "score"]), uniform=True)

        combined = mock1 + mock2
        # we dont need an input, as both Identity transformers will return anyway
        rtr = combined.transform(None)

        self.assertEqual(1, len(rtr))
        self.assertEqual("q1", rtr.iloc[0]["qid"])
        self.assertEqual("doc1", rtr.iloc[0]["docno"])
        self.assertEqual(15, rtr.iloc[0]["score"])

    def test_plus_notoverlap(self):
        import pyterrier.transformer as ptt
        mock1 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 5]], columns=["qid", "docno", "score"]), uniform=True)
        mock2 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc2", 10]], columns=["qid", "docno", "score"]), uniform=True)

        combined = mock1 + mock2
        # we dont need an input, as both Identity transformers will return anyway
        rtr = combined.transform(None)

        self.assertEqual(2, len(rtr))
        self.assertEqual("q1", rtr.iloc[0]["qid"])
        self.assertEqual("doc1", rtr.iloc[0]["docno"])
        self.assertEqual(5, rtr.iloc[0]["score"])
        self.assertEqual("doc2", rtr.iloc[1]["docno"])
        self.assertEqual(10, rtr.iloc[1]["score"])

    def test_plus_more_cols(self):
        from pyterrier.model import add_ranks
        mock1 = pt.Transformer.from_df(add_ranks(pd.DataFrame([["q1", "a query", "doc1", 5]], columns=["qid", "query", "docno", "score"]), single_query=True), uniform=True)
        mock2 = pt.Transformer.from_df(add_ranks(pd.DataFrame([["q1", "a query", "doc1", 10]], columns=["qid", "query", "docno", "score"]), single_query=True), uniform=True)

        combined = mock1 + mock2
        # we dont need an input, as both Identity transformers will return anyway
        rtr = combined.transform(None)

        self.assertEqual(1, len(rtr))
        self.assertEqual("q1", rtr.iloc[0]["qid"])
        self.assertEqual("doc1", rtr.iloc[0]["docno"])
        self.assertEqual(15, rtr.iloc[0]["score"])
        bad_columns = ["rank_x", "rank_y", "rank_r", "query_x", "query_y", "query_R", "score_x", "score_y", "score_r"]
        for bad in bad_columns:
            self.assertFalse(bad in rtr.columns, "column %s in returned dataframe" % bad)

    def test_rank_cutoff(self):
        mock1 = pt.Transformer.from_df( pd.DataFrame([["q1", "d2", 0, 5.1], ["q1", "d3", 1, 5.1]], columns=["qid", "docno", "rank", "score"]), uniform=True)
        cutpipe = mock1 % 1
        rtr = cutpipe.transform(None)
        self.assertEqual(1, len(rtr))
        
    def test_concatenate(self):
        import numpy as np
        mock1 = pt.Transformer.from_df( pd.DataFrame([["q1", "d2", 2, 4.9, np.array([1,2])], ["q1", "d3", 1, 5.1, np.array([1,2])]], columns=["qid", "docno", "rank", "score", "bla"]), uniform=True)
        mock2 = pt.Transformer.from_df( pd.DataFrame([["q1", "d1", 1, 4.9, np.array([1,1])], ["q1", "d3", 2, 5.1, np.array([1,2])]], columns=["qid", "docno", "rank", "score", "bla"]), uniform=True)

        cutpipe = mock1 ^ mock2
        rtr = cutpipe.transform(None)
        self.assertEqual(3, len(rtr))
        row0 = rtr.iloc[0] 
        self.assertEqual("d3", row0["docno"])
        self.assertEqual(5.1, row0["score"])
        row1 = rtr.iloc[1] 
        self.assertEqual("d2", row1["docno"])
        self.assertEqual(4.9, row1["score"])
        row2 = rtr.iloc[2] 
        self.assertEqual("d1", row2["docno"])
        self.assertEqual(4.9-0.0001, row2["score"])


    def test_plus_multi_rewrite(self):
        mock1 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 5]], columns=["qid", "docno", "score"]), uniform=True)
        mock2 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 10]], columns=["qid", "docno", "score"]), uniform=True)
        mock3 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 15]], columns=["qid", "docno", "score"]), uniform=True)

        combined = mock1 + mock2 + mock3
        for pipe in [combined, combined.compile()]:

            # we dont need an input, as both Identity transformers will return anyway
            rtr = pipe.transform(None)

            self.assertEqual(1, len(rtr))
            self.assertEqual("q1", rtr.iloc[0]["qid"])
            self.assertEqual("doc1", rtr.iloc[0]["docno"])
            self.assertEqual(30, rtr.iloc[0]["score"])


    def test_union(self):
        mock1 = pt.Transformer.from_df(pd.DataFrame([["q1", "q1texta", "doc1", 5, "body text"], ["q1", "q1texta", "doc3", 5, "body text"]], columns=["qid", "query", "docno", "score", "body"]), uniform=True)
        mock2 = pt.Transformer.from_df(pd.DataFrame([["q1", "q1textb", "doc2", 10, "body text" ]], [["q1", "q1textb", "doc3", 10, "body text"]], columns=["qid", "query", "docno", "score", "body"]), uniform=True)

        combined = mock1 | mock2
        # we dont need an input, as both Identity transformers will return anyway
        rtr = combined.transform(None)

        self.assertEqual(3, len(rtr))
        self.assertTrue("q1" in rtr["qid"].values)
        self.assertTrue("doc1" in rtr["docno"].values)
        self.assertTrue("doc2" in rtr["docno"].values)
        # in case we have different values for query for the same (qid, docno), we use only the first one
        self.assertTrue("q1texta" in rtr["query"].values)
        self.assertTrue("q1textb" in rtr[rtr.docno == "doc2"]["query"].values) 
        self.assertTrue("q1textb" not in rtr[rtr.docno == "doc3"]["query"].values) 

        print(rtr.columns)
        for col in ["qid", "query", "docno", "body"]:
            self.assertTrue(col in rtr.columns, "%s not found in cols" % col)
        for col in ["rank", "score", 'query_y', 'score_y', 'body_y']:
            self.assertFalse(col in rtr.columns, "%s found in cols" % col)            

    def test_intersect(self):
        mock1 = pt.Transformer.from_df(pd.DataFrame([["q1", "q1texta", "doc1", 5, "body text"]], columns=["qid", "query", "docno", "score", "body"]), uniform=True)
        mock2 = pt.Transformer.from_df(pd.DataFrame([["q1", "q1textb", "doc2", 10, "body text"], ["q1", "q1textb", "doc1", 10, "body text"]], columns=["qid", "query", "docno", "score", "body"]), uniform=True)

        combined = mock1 & mock2
        # we dont need an input, as both Identity transformers will return anyway
        rtr = combined.transform(None)

        self.assertEqual(1, len(rtr))
        self.assertTrue("q1" in rtr["qid"].values)
        self.assertTrue("doc1" in rtr["docno"].values)
        self.assertFalse("doc2" in rtr["docno"].values)
        # in case we have different values for query for the same (qid, docno), we use the left one
        self.assertTrue("q1texta" in rtr["query"].values)
        self.assertTrue("q1textb" not in rtr["query"].values)

        for col in ["qid", "query", "docno", "body"]:
            self.assertTrue(col in rtr.columns, "%s not found in cols" % col)

        for col in ["rank", "score", 'query_y', 'score_y', 'body_y']:
            self.assertFalse(col in rtr.columns, "%s found in cols" % col)        
        
    def test_feature_union_multi_actual(self):
        dataset = pt.get_dataset("vaswani")
        index = dataset.get_index()
        BM25 = pt.terrier.Retriever(index, wmodel="BM25")
        TF_IDF = pt.terrier.Retriever(index, wmodel="TF_IDF")
        PL2 = pt.terrier.Retriever(index, wmodel="PL2")
        expression = BM25 >> (pt.transformer.IdentityTransformer() ** TF_IDF ** PL2)

        self.assertEqual(2, len(expression))
        self.assertEqual(3, len(expression[1]))

        res = expression.transform(dataset.get_topics().head(2))
        self.assertTrue("features" in res.columns)
        self.assertFalse("features_x" in res.columns)
        self.assertFalse("features_y" in res.columns)
        print(res.iloc[0]["features"])
        self.assertEqual(3, len(res.iloc[0]["features"]))


    def test_feature_union_multi(self):
        import pyterrier._ops as pto
        mock0 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 0], ["q1", "doc2", 0]], columns=["qid", "docno", "score"]), uniform=True)

        mock1 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 5], ["q1", "doc2", 0]], columns=["qid", "docno", "score"]), uniform=True)
        mock2 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 10], ["q1", "doc2", 0]], columns=["qid", "docno", "score"]), uniform=True)
        mock3 = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 15], ["q1", "doc2", 0]], columns=["qid", "docno", "score"]), uniform=True)

        mock3_empty = pt.Transformer.from_df(pd.DataFrame([], columns=["qid", "docno", "score"]), uniform=True)
        mock2_partial = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 10]], columns=["qid", "docno", "score"]), uniform=True)
        mock3_partial = pt.Transformer.from_df(pd.DataFrame([["q1", "doc1", 15]], columns=["qid", "docno", "score"]), uniform=True)


        mock12a = mock1 ** mock2
        mock123a = mock1 ** mock2 ** mock3
        mock123b = mock12a ** mock3
        mock123a_manual = pto.FeatureUnion(
                pto.FeatureUnion(mock1, mock2),
                mock3
        )
        mock123b_manual = pto.FeatureUnion(
                mock1,
                pto.FeatureUnion(mock2, mock3),
        )
        mock123e = pto.FeatureUnion(
                mock1,
                pto.FeatureUnion(mock2, mock3_empty),
        )

        mock12e3 = pto.FeatureUnion(
                mock1,
                pto.FeatureUnion(mock3_empty, mock3),
        )

        mock123p = pto.FeatureUnion(
                mock1,
                pto.FeatureUnion(mock2, mock3_partial),
        )

        mock12p3 = pto.FeatureUnion(
                mock1,
                pto.FeatureUnion(mock2_partial, mock3),
        )
        
        
        self.assertEqual(2, len(mock12a))
        self.assertEqual(2, len(mock12a))

        mock123_simple = mock123a.compile()
        self.assertIsNotNone(mock123_simple)
        self.assertEqual("(UniformTransformer() ** UniformTransformer() ** UniformTransformer())", repr(mock123_simple))
        #
        #
        self.assertEqual(3, len(mock123_simple))

        def _test_expression(expression):
            # we dont need an input, as both Identity transformers will return anyway
            rtr = (mock0 >> expression).transform(None)
            #print(rtr)
            self.assertIsNotNone(rtr)
            self.assertEqual(2, len(rtr))
            self.assertTrue("qid" in rtr.columns)
            self.assertTrue("docno" in rtr.columns)
            self.assertFalse("features_x" in rtr.columns)
            self.assertFalse("features_y" in rtr.columns)
            self.assertTrue("features" in rtr.columns)
            self.assertTrue("q1" in rtr["qid"].values)
            self.assertTrue("doc1" in rtr["docno"].values)
            import numpy as np
            self.assertTrue( np.allclose(np.array([5,10,15]), rtr.iloc[0]["features"]))

        _test_expression(mock123_simple)
        _test_expression(mock123a)
        _test_expression(mock123b)
        _test_expression(mock123b)
        with self.assertRaises(ValueError):
            _test_expression(mock123e)
        with self.assertRaises(ValueError):
            _test_expression(mock12e3)
        
        with warnings.catch_warnings(record=True) as w:
            _test_expression(mock123p)
            assert "Got number of results" in str(w[-1].message)
        
        with warnings.catch_warnings(record=True) as w:
            _test_expression(mock12p3)
            assert "Got number of results" in str(w[-1].message)

    def test_feature_union_empty(self):
        mock_input = pt.Transformer.from_df(pd.DataFrame([["q1", "a query", "doc1", 5]], columns=["qid", "query", "docno", "score"]))
        mock_f1 = pt.Transformer.from_df(pd.DataFrame([["q1", "a query", "doc1", 10]], columns=["qid", "query", "docno", "score"]))
        mock_f2 = pt.Transformer.from_df(pd.DataFrame([["q1", "a query", "doc1", 50]], columns=["qid", "query", "docno", "score"]))

        pipe = mock_input >> (mock_f1 ** mock_f2)
        rtr = pipe.search('another query', qid='q2')
        self.assertEqual(0, len(rtr))
        self.assertIn('score', rtr.columns)
        self.assertIn('features', rtr.columns)

    def test_feature_union(self): 
        import pyterrier._ops as ptt
        mock_input = pt.Transformer.from_df(pd.DataFrame([["q1", "a query", "doc1", 5]], columns=["qid", "query", "docno", "score"]), uniform=True)
        
        mock_f1 = pt.Transformer.from_df(pd.DataFrame([["q1", "a query", "doc1", 10]], columns=["qid", "query", "docno", "score"]), uniform=True)
        mock_f2 = pt.Transformer.from_df(pd.DataFrame([["q1", "a query", "doc1", 50]], columns=["qid", "query", "docno", "score"]), uniform=True)

        def _test_expression(pipeline):
            # check access to the objects
            self.assertEqual(2, len(pipeline))
            self.assertEqual(2, len(pipeline[1]))
            
            # we dont need an input, as both Uniform transformers will return anyway
            for name, input in [
                ('none', None),
                ('query df', pt.new.queries(['a query'], qid=['q1'])),
                ('empty query df', pt.new.queries(['a query'], qid=['q1']).head(0))
            ]:
                with self.subTest(name):
                    rtr = pipeline.transform(input)
                    self.assertEqual(1, len(rtr))
                    self.assertTrue("qid" in rtr.columns)
                    self.assertTrue("docno" in rtr.columns)
                    #self.assertTrue("score" in rtr.columns)
                    self.assertTrue("features" in rtr.columns)

                    bad_columns = ["rank_x", "rank_y", "rank_r", "query_x", "query_y", "query_R", "score_x", "score_y", "score_r", "features_x", "features_y"]
                    for bad in bad_columns:
                        self.assertFalse(bad in rtr.columns, "column %s in returned dataframe" % bad)

                    self.assertTrue("q1" in rtr["qid"].values)
                    self.assertTrue("doc1" in rtr["docno"].values)
                    import numpy as np
                    self.assertTrue( np.array_equal(np.array([10,50]), rtr.iloc[0]["features"]))

        # test using direct instantiation, as well as using the ** operator
        _test_expression(mock_input >> ptt.FeatureUnion(mock_f1, mock_f2))
        _test_expression(mock_input >> mock_f1 ** mock_f2)  

    def _assertInNoOrder(self, member, container):
        container_set = [set(item) for item in container]
        self.assertIn(set(member), container_set)     

    def test_funion_inspection(self):
        dataset = pt.get_dataset("vaswani")
        index = dataset.get_index()
        BM25 = pt.terrier.Retriever(index, wmodel="BM25")
        pipe = (BM25 ** BM25)
        self._assertInNoOrder(['qid', 'docno', 'query'], pt.inspect.transformer_inputs(pipe)) 


if __name__ == "__main__":
    unittest.main()
        



import pyterrier as pt
from pyterrier.model import coerce_dataframe_types
from ._rendering import _convert_measures
from . import MEASURES_TYPE, SYSTEM_OR_RESULTS_TYPE, SAVEMODE_TYPE, SAVEFORMAT_TYPE

from .. import Transformer
from .._ops import Compose

import ir_measures
import pandas as pd
import tqdm as tqdm_module
from ir_measures import Measure, Metric
import os
import sys
import types
from typing import Any, Dict, Iterator, List, Optional, Sequence, Tuple, Union
from warnings import warn

#list(iter_calc([ir_measures.AP], qrels, run))
#[Metric(query_id='Q0', measure=AP, value=1.0), Metric(query_id='Q1', measure=AP, value=1.0)]
def _ir_measures_to_dict(
        seq : Iterator[Metric], 
        measures: Sequence[Measure],
        rev_mapping : Dict[Measure,str], 
        num_q : int,
        perquery : bool = True,
        backfill_qids : Optional[Sequence[str]] = None) -> Union[ Dict[str, Dict[str, float]], Dict[str, float]]:
    from collections import defaultdict
    if perquery:
        # qid -> measure -> value
        rtr_perquery : Dict[str, Dict[str, float]] = defaultdict(dict)
        for metric in seq:
            measure = metric.measure
            measure_name = rev_mapping.get(measure, str(measure))
            rtr_perquery[metric.query_id][measure_name] = metric.value
        # When reporting per-query results, it can desirable to show something for topics that were executed
        # do not have corresponding qrels. If the caller passes in backfill_qids, we'll ensure that these
        # qids are present, and if not add placeholders with NaN values for all measures.
        if backfill_qids is not None:
            backfill_count = 0
            for qid in backfill_qids:
                if qid not in rtr_perquery:
                    backfill_count += 1
                    for m in measures:
                        rtr_perquery[qid][rev_mapping.get(m, str(m))] = float('NaN')
            if backfill_count > 0:
                warn(f'{backfill_count} topic(s) not found in qrels. Scores for these topics are given as NaN and should not contribute to averages.')
        return rtr_perquery
    assert backfill_qids is None, "backfill_qids only supported when perquery=True"
    
    metric_agg = {rev_mapping.get(m, str(m)): m.aggregator() for m in measures}
    for metric in seq:
        measure_name = rev_mapping.get(metric.measure, str(metric.measure))
        metric_agg[measure_name].add(metric.value)
    
    rtr_aggregated : Dict[str,float] = {} # measure -> value
    for m_name in metric_agg:
        rtr_aggregated[m_name] = metric_agg[m_name].result()
    return rtr_aggregated

def _validate_R_is_res(df : pd.DataFrame):
    found = []
    unfound = []
    for c in ["qid", "docno", "score", "rank"]:
        if c in df.columns:
            found.append(c)
        else:
            unfound.append(c)
    if len(unfound):
        raise TypeError("save_dir was set, but results dont look like R (expected and found %s, missing %s). You probably need to set save_format kwarg, "
                        "e.g. save_format=pickle" %
                        (str(found), str(unfound)))


def _run_and_evaluate(
        system : SYSTEM_OR_RESULTS_TYPE,
        topics : Optional[pd.DataFrame],
        qrels: pd.DataFrame,
        metrics : MEASURES_TYPE,
        pbar : Optional[tqdm_module.tqdm] = None,
        save_mode : Optional[SAVEMODE_TYPE] = None,
        save_file : Optional[str] = None,
        save_format : SAVEFORMAT_TYPE = 'trec',
        perquery : bool = False,
        batch_size : Optional[int] = None,
        backfill_qids : Optional[Sequence[str]] = None):

    from ..io import read_results, write_results

    if pbar is None:
        pbar = pt.tqdm(disable=True)

    metrics, rev_mapping = _convert_measures(metrics)
    qrels = pt.model.to_ir_measures(qrels)
    from timeit import default_timer as timer
    runtime : float = 0.
    num_q = qrels['query_id'].nunique()
    if save_file is not None and os.path.exists(save_file):
        if save_mode == 'reuse':
            if save_format == 'trec':
                system = read_results(save_file)
            elif isinstance(save_format, types.ModuleType):
                with pt.io.autoopen(save_file, 'rb') as fin:
                    system = save_format.load(fin)
            elif isinstance(save_format, tuple) and len(save_format) == 2:
                with pt.io.autoopen(save_file, 'rb') as fin:
                    system = save_format[0](fin)
            else:
                raise ValueError("Unknown save_format %s" % str(save_format))
        elif save_mode == 'overwrite':
            os.remove(save_file)
        elif save_mode == 'warn':
            warn(("save_dir is set, but the file '%s' already exists. If you are aware of are happy to reuse this " % save_file)+
                             "file to speed up evaluation, set save_mode='reuse'; if you want to overwrite it, set save_mode='overwrite'."+
                             " To make this condition an error, use save_mode='error'.")
        elif save_mode == 'error':
            raise ValueError(("save_dir is set, but the file '%s' already exists. If you are aware of are happy to reuse this " % save_file)+
                             "file to speed up evaluation, set save_mode='reuse'; if you want to overwrite it, set save_mode='overwrite'."+
                              "To make this condition a warning, use save_mode='warn'.")
        else:
            raise ValueError("Unknown save_mode argument '%s', valid options are 'error', 'warn', 'reuse' or 'overwrite'." % save_mode)

    res : pd.DataFrame
    # if its a DataFrame, use it as the results
    if isinstance(system, pd.DataFrame):
        res = system
        res = coerce_dataframe_types(res)
        if len(res) == 0:
            if topics is None:
                raise ValueError("No topics specified, and no results in dataframe")
            else:
                raise ValueError("%d topics, but no results in dataframe" % len(topics))
        evalMeasuresDict = _ir_measures_to_dict(
            ir_measures.iter_calc(metrics, qrels, pt.model.to_ir_measures(res)),
            metrics,
            rev_mapping,
            num_q,
            perquery,
            backfill_qids)
        pbar.update()

    elif batch_size is None:

        assert topics is not None, "topics must be specified"
        #transformer, evaluate all queries at once

        starttime = timer()
        res = system.transform(topics)
        endtime = timer()
        runtime =  float(endtime - starttime) * 1000.

        # write results to save_file; we can be sure this file does not exist
        if save_file is not None:
            if save_format == 'trec':
                _validate_R_is_res(res)
                write_results(res, save_file)
            elif isinstance(save_format, types.ModuleType):
                with pt.io.autoopen(save_file, 'wb') as fout:
                    save_format.dump(res, fout)
            elif isinstance(save_format, tuple) and len(save_format) == 2:
                with pt.io.autoopen(save_file, 'wb') as fout:
                    save_format[1](res, fout)
            else:
                raise ValueError("Unknown save_format %s" % str(save_format))
        res = coerce_dataframe_types(res)

        if len(res) == 0:
            raise ValueError("%d topics, but no results received from %s" % (len(topics), str(system)) )

        evalMeasuresDict = _ir_measures_to_dict(
            ir_measures.iter_calc(metrics, qrels, pt.model.to_ir_measures(res)),
            metrics,
            rev_mapping,
            num_q,
            perquery,
            backfill_qids)
        pbar.update()
    else:
        assert topics is not None, "topics must be specified"
        if save_file is not None:
            # only 
            assert save_format == 'trec', 'save_format=%s is not supported when save_dir is enabled and batch_size is not None' % str(save_format)

        #transformer, evaluate queries in batches
        assert batch_size > 0
        starttime = timer()
        evalMeasuresDict = {}
        remaining_qrel_qids = set(qrels.query_id)
        try:
            batch_topics : pd.DataFrame
            for i, (res, batch_topics) in enumerate( system.transform_gen(topics, batch_size=batch_size, output_topics=True)):
                if len(res) == 0:
                    raise ValueError("batch of %d topics, but no results received in batch %d from %s" % (len(batch_topics), i, str(system) ) )
                endtime = timer()
                runtime += (endtime - starttime) * 1000.

                # write results to save_file; we will append for subsequent batches
                if save_file is not None:
                    _validate_R_is_res(res)
                    write_results(res, save_file, append=True)

                res = coerce_dataframe_types(res)
                batch_qids = set(batch_topics.qid)
                batch_qrels = qrels[qrels.query_id.isin(batch_qids)] # filter qrels down to just the qids that appear in this batch
                remaining_qrel_qids.difference_update(batch_qids)
                batch_backfill = [qid for qid in backfill_qids if qid in batch_qids] if backfill_qids is not None else None
                evalMeasuresDict.update(_ir_measures_to_dict( # type: ignore[arg-type]
                    ir_measures.iter_calc(metrics, batch_qrels, pt.model.to_ir_measures(res)),
                    metrics,
                    rev_mapping,
                    num_q,
                    perquery=True,
                    backfill_qids=batch_backfill))
                pbar.update()
                starttime = timer()
        except:
            # if an error is thrown, we need to clean up our existing file
            if save_file is not None and os.path.exists(save_file):
                os.remove(save_file)
            raise
        if remaining_qrel_qids:
            # there are some qids in the qrels that were not in the topics. Get the default values for these and update evalMeasuresDict
            missing_qrels = qrels[qrels.query_id.isin(remaining_qrel_qids)]
            empty_res = pd.DataFrame([], columns=['query_id', 'doc_id', 'score'])
            evalMeasuresDict.update(_ir_measures_to_dict( # type: ignore[arg-type]
                ir_measures.iter_calc(metrics, missing_qrels, empty_res),
                metrics,
                rev_mapping,
                num_q,
                perquery=True))
        if not perquery:
            # aggregate measures if not in per query mode
            aggregators: Dict[str, Any] = {rev_mapping.get(m, str(m)): m.aggregator() for m in metrics}
            q : str
            for q in evalMeasuresDict:
                for metric in metrics:
                    s_metric = rev_mapping.get(metric, str(metric))
                    aggregators[s_metric].add(evalMeasuresDict[q][s_metric]) #type: ignore
            evalMeasuresDict = {m: agg.result() for m, agg in aggregators.items()}
    return (runtime, evalMeasuresDict)


def _identifyCommon(pipes : List[Union[pt.Transformer, pd.DataFrame]]) -> Tuple[Optional[pt.Transformer], List[Union[pt.Transformer,pd.DataFrame]]]:
    # constructs a common prefix pipeline across a list of pipelines, along with various suffices. 
    # pt.Transformer.identity() is used for a no-op suffix

    # no precomputation for single-system case
    if len(pipes) == 1:
        return None, pipes
    pipe_lists: List[List[pt.Transformer]] = []
    for p in pipes:
        # no optimisation possible for experiments involving dataframes as systems
        if isinstance(p, pd.DataFrame):
            return None, pipes
        if isinstance(p, Compose):
            pipe_lists.append(list(p._transformers))
        else:
            if not isinstance(p, pt.Transformer):
                raise ValueError("pt.Experiment has systems that are not either DataFrames or Transformers")
            pipe_lists.append([p])

    common_prefix, suffices  = _common_prefix(pipe_lists)

    if len(common_prefix) == 0:
        # no common prefix, return existing pipelines as-is
        return None, pipes

    def _construct(inp: List[pt.Transformer]) -> pt.Transformer:
        # use identify as a no-op
        if len(inp) == 0:
            return pt.Transformer.identity()
        # use transformer itself
        if len(inp) == 1:
            return inp[0]
        # more than 1, compose...
        return Compose(*inp)

    return (
        _construct(common_prefix), # prefix common to all 
        [ _construct(remainder) for remainder in suffices ] # individual suffices
    )


def _precomputation(
        retr_systems : List[SYSTEM_OR_RESULTS_TYPE],
        topics : pd.DataFrame,
        precompute_prefix : bool,
        verbose : bool,
        batch_size : Optional[int] = None
        ) -> Tuple[float, pd.DataFrame, List[SYSTEM_OR_RESULTS_TYPE]]:

    # this method identifies any common prefix to the pipelines in retr_systems,
    # and if precompute_prefix=True, then it computes the (partial) results of topics
    # on that prefix, and returns that, which is used later for evaluating the remainder
    # of each pipeline.

    tqdm_args_precompute: Dict[str, Any] = {
        'disable' : not verbose,
    }

    common_pipe, execution_retr_systems = _identifyCommon(retr_systems)
    precompute_time = 0.
    if precompute_prefix and common_pipe is not None:
        print("Precomputing results of %d topics on shared pipeline component %s" % (len(topics), str(common_pipe)), file=sys.stderr)

        tqdm_args_precompute['desc'] = "pt.Experiment precomputation"
        from timeit import default_timer as timer
        starttime = timer()
        if batch_size is not None:

            warn("precompute_prefix with batch_size is very experimental. Please report any problems")
            import math
            tqdm_args_precompute['unit'] = 'batches'
            # round number of batches up for each system
            tqdm_args_precompute['total'] = math.ceil((len(topics) / batch_size))
            with pt.tqdm(**tqdm_args_precompute) as pbar:
                precompute_results : List[pd.DataFrame] = []
                for r in common_pipe.transform_gen(topics, batch_size=batch_size):
                    assert isinstance(r, pd.DataFrame) # keep mypy happy
                    precompute_results.append(r)
                    pbar.update(1)
                execution_topics = pd.concat(precompute_results)

        else: # no batching  
            tqdm_args_precompute['total'] = 1
            tqdm_args_precompute['unit'] = "prefix pipeline"
            with pt.tqdm(**tqdm_args_precompute) as pbar:
                execution_topics = common_pipe(topics)
                pbar.update(1)

        endtime = timer()
        precompute_time = float(endtime - starttime) * 1000.

    elif precompute_prefix and common_pipe is None:
        warn('precompute_prefix was True for pt.Experiment, but no common pipeline prefix was found among %d pipelines' % len(retr_systems))
        execution_retr_systems = retr_systems
        execution_topics = topics

    else: # precomputation not requested
        execution_retr_systems = retr_systems
        execution_topics = topics

    return precompute_time, execution_topics, execution_retr_systems


def _common_prefix(pipes: List[List[Transformer]]) -> Tuple[List[Transformer], List[List[Transformer]]]:
    # finds the common prefix within a list of transformers
    assert len(pipes) > 1, "pipes must contain at least two lists"
    common_prefix = []
    for stage in zip(*pipes):
        elements = set(stage) # uses Transformer.__equals__ 
        if len(elements) == 1:
            common_prefix.append(next(iter(elements)))
        else:
            break
    suffixes = [p[len(common_prefix):] for p in pipes]
    return common_prefix, suffixes


from typing import Union, List
import pyterrier as pt


@pt.java.required
class IndexFactory:
    """
    The ``of()`` method of this factory class allows to load a Terrier `Index <http://terrier.org/docs/current/javadoc/org/terrier/structures/Index.html>`_.

    NB: This class "shades" the native Terrier `IndexFactory <http://terrier.org/docs/current/javadoc/org/terrier/structures/IndexFactory.html>`_ class - it offers essential the same API,
    except that the ``of()`` method contains a memory kwarg, that can be used to load additional index data structures into memory. 

    Terrier data structures that can be loaded into memory:
     - 'inverted' - the inverted index, contains posting lists for each term. In the default configuration, this is read in from disk in chunks.
     - 'lexicon' - the dictionary. By default, a binary search of the on-disk structure is used, so loading into memory can enhance speed.
     - 'meta' - metadata about documents. Used as the final stage of retrieval, one seek for each retrieved document.
     - 'direct' - contains posting lists for each document. No speed advantage for loading into memory unless pseudo-relevance feedback is being used.
     - 'document' - contains document lengths, which are anyway loaded into memory. No speed advantage for loading into memory unless pseudo-relevance feedback is being used.
    """

    @staticmethod
    def _load_into_memory(index, structures=['lexicon', 'direct', 'inverted', 'meta'], load=False):

        REWRITES = {
            'meta' : {
                # both metaindex implementations have the same property
                'org.terrier.structures.ZstdCompressedMetaIndex' : {
                    'index.meta.index-source' : 'fileinmem',
                    'index.meta.data-source' : 'fileinmem'},
            
                'org.terrier.structures.CompressingMetaIndex' : {
                    'index.meta.index-source' : 'fileinmem',
                    'index.meta.data-source' : 'fileinmem'}
            },
            'lexicon' : {
                'org.terrier.structures.FSOMapFileLexicon' : {
                    'index.lexicon.data-source' : 'fileinmem'
                }
            },
            'direct' : {
                'org.terrier.structures.bit.BitPostingIndex' : {
                    'index.direct.data-source' : 'fileinmem'}
            },
            'inverted' : {
                'org.terrier.structures.bit.BitPostingIndex' : {
                    'index.inverted.data-source' : 'fileinmem'}
            },
        }
        if "direct" in structures:
            REWRITES['document'] = {
                # we have to be sensitive to the presence of fields or not
                # NB: loading these structures into memory only benefit direct index access
                'org.terrier.structures.FSADocumentIndex' : {
                    'index.document.class' : 'FSADocumentIndexInMem'
                }, 
                'org.terrier.structures.FSAFieldDocumentIndex' : {
                    'index.document.class' : 'FSADocumentIndexInMemFields'
                }
            }

        pindex = pt.java.cast("org.terrier.structures.IndexOnDisk", index)
        dirty_structures = set()
        for s in structures:
            if not pindex.hasIndexStructure(s):
                continue
            clz = pindex.getIndexProperty(f"index.{s}.class", "notfound")
            if clz not in REWRITES[s]:
                raise ValueError(f"Cannot load structure {s} into memory, underlying class {clz} is not supported")

            # we only reload an index structure if a property has changed
            for k, v in REWRITES[s][clz].items():
                if pindex.getIndexProperty(k, "notset") != v:
                    pindex.setIndexProperty(k, v)
                    dirty_structures.add(s)

                    # if the document index is reloaded, the inverted index should be reloaded too
                    # NB: the direct index needs reloaded too, but this option is only available IF
                    # the direct index is setup
                    if s == "document":
                        dirty_structures.add("inverted")

        # remove the old data structures from memory
        for s in dirty_structures:
            if pindex.structureCache.containsKey(s):
                pindex.structureCache.remove(s)

        # force the index structures to be loaded now
        if load:
            for s in dirty_structures:
                pindex.getIndexStructure(s)

        # dont allow the index properties to be rewritten
        pindex.dirtyProperties = False
        return index

    @staticmethod 
    def of(indexlike, memory : Union[bool, List[str]] = False):
        """
        Loads an index. Returns a Terrier `Index <http://terrier.org/docs/current/javadoc/org/terrier/structures/Index.html>`_ object.

        :param indexlike: The location of the index. This can be a string, or an `IndexRef <http://terrier.org/docs/current/javadoc/org/terrier/querying/IndexRef.html>`__ object.
        :param memory: If the index should be loaded into memory. Use `True` for all structures, or a list of structure names.
        :return: A Terrier `Index <http://terrier.org/docs/current/javadoc/org/terrier/structures/Index.html>`_ object.
        """
        load_profile = pt.terrier.J.IndexOnDisk.getIndexLoadingProfileAsRetrieval()

        if memory or (isinstance(memory, list) and len(memory) > 0): #MEMORY CAN BE A LIST?
            pt.terrier.J.IndexOnDisk.setIndexLoadingProfileAsRetrieval(False)
        index = pt.terrier.J.IndexFactory.of(indexlike)
        
        # noop if memory is False
        pt.terrier.J.IndexOnDisk.setIndexLoadingProfileAsRetrieval(load_profile)
        if not memory:
            return index
        if isinstance(memory, list):
            return IndexFactory._load_into_memory(index, structures=memory)
        return IndexFactory._load_into_memory(index)



"""
This file contains all the indexers.
"""
from enum import Enum
import pandas as pd
import os
import json
import tempfile
import contextlib
import threading
import select
import math
from warnings import warn
from deprecated import deprecated
from collections import deque
from typing import List, Dict, Union, Any, Callable, Type
import more_itertools
import pyterrier as pt
from pyterrier.terrier.stemmer import TerrierStemmer
from pyterrier.terrier.tokeniser import TerrierTokeniser
from pyterrier.terrier.stopwords import TerrierStopwords


# These classes are only defined after pt.java.init() in pyterrier.terrier.java._post_init
DocListIterator: Any = None
PythonListIterator: Any = None
FlatJSONDocumentIterator: Any = None
TQDMCollection: Any = None
TQDMSizeCollection: Any = None

# lastdoc ensures that a Document instance from a Collection is not GCd before Java has used it.
lastdoc = None

type_to_class = {
    'trec' : 'org.terrier.indexing.TRECCollection',
    'trecweb' : 'org.terrier.indexing.TRECWebCollection',
    'warc' : 'org.terrier.indexing.WARC10Collection'
}

@pt.java.required
def createCollection(files_path : List[str], coll_type : str = 'trec', props = {}):
    if coll_type in type_to_class:
        collectionClzName = type_to_class[coll_type]
    else:
        collectionClzName = coll_type
    _props = pt.java.J.HashMap()
    for k, v in props.items():
        _props[k] = v
    pt.terrier.J.ApplicationSetup.getProperties().putAll(_props)
    cls_string = pt.java.J.String._class
    cls_list = pt.java.J.List._class
    if len(files_path) == 0:
        raise ValueError("list files_path cannot be empty")
    asList = createAsList(files_path)
    colObj = pt.terrier.J.CollectionFactory.loadCollections(
        collectionClzName.split(","),
        [cls_list, cls_string, cls_string, cls_string],
        [asList, pt.terrier.J.TagSet.TREC_DOC_TAGS, "", ""])
    return colObj


def treccollection2textgen(
        files : List[str], 
        meta : List[str] = ["docno"], 
        meta_tags : Dict[str,str] = {"text":"ELSE"}, 
        verbose = False,
        num_docs = None,
        tag_text_length : int = 4096):
    """
    Creates a generator of dictionaries on parsing TREC formatted files. This is useful 
    for parsing TREC-formatted corpora in indexers like IterDictIndexer, or similar 
    indexers in other plugins (e.g. ColBERTIndexer).

    "Arguments:
     -" files(List[str]): list of files to parse in TREC format.
     - meta(List[str]): list of attributes to expose in the dictionaries as metadata.
     - meta_tags(Dict[str,str]): mapping of TREC tags as metadata.
     - tag_text_length(int): maximium length of metadata. Defaults to 4096.
     - verbose(bool): set to true to show a TQDM progress bar. Defaults to True.
     - num_docs(int): a hint for TQDM to size the progress bar based on document counts rather than file count.

    Example::

        files = pt.io.find_files("/path/to/Disk45")
        gen = pt.index.treccollection2textgen(files)
        index = pt.IterDictIndexer("./index45").index(gen)

    """

    props = {
        "TrecDocTags.doctag": "DOC",
        "TrecDocTags.idtag": "DOCNO",
        "TrecDocTags.skip": "DOCHDR",
        "TrecDocTags.casesensitive": "false",
        # Should the tags from which we create abstracts be case-sensitive?
        'TaggedDocument.abstracts.tags.casesensitive':'false'
    }
    props['TaggedDocument.abstracts'] = ','.join(meta_tags.keys())
    # The tags from which to save the text. ELSE is special tag name, which means anything not consumed by other tags.
    props['TaggedDocument.abstracts.tags'] = ','.join(meta_tags.values())
    # The max lengths of the abstracts. Abstracts will be truncated to this length. Defaults to empty.
    props['TaggedDocument.abstracts.lengths'] = ','.join([str(tag_text_length)] * len(meta_tags) )  

    collection = createCollection(files, props=props)
    if verbose:
        if num_docs is not None:
            collection = TQDMSizeCollection(collection, num_docs)
        else:
            collection = TQDMCollection(collection)
    while collection.nextDocument():
        d = collection.getDocument()
        while not d.endOfDocument():
            d.getNextTerm()
        rtr = {}
        for k in meta:
            rtr[k] = d.getProperty(k)
        for k in meta_tags:
            rtr[k] = d.getProperty(k)
        yield rtr
    

@pt.java.required
def _TaggedDocumentSetup(
        meta : Dict[str,int],  #mapping from meta-key to length
        meta_tags : Dict[str,str] #mapping from meta-key to tag
    ):
    """
    Property setup for TaggedDocument etc to generate abstracts
    """

    abstract_tags=meta_tags.values()
    abstract_names=meta_tags.keys()
    abstract_lengths=[str(meta[name]) for name in abstract_names]

    pt.terrier.J.ApplicationSetup.setProperty("TaggedDocument.abstracts", ",".join(abstract_names))
    # The tags from which to save the text. ELSE is special tag name, which means anything not consumed by other tags.
    pt.terrier.J.ApplicationSetup.setProperty("TaggedDocument.abstracts.tags", ",".join(abstract_tags))
    # The max lengths of the abstracts. Abstracts will be truncated to this length. Defaults to empty.
    pt.terrier.J.ApplicationSetup.setProperty("TaggedDocument.abstracts.lengths", ",".join(abstract_lengths))
    # Should the tags from which we create abstracts be case-sensitive
    pt.terrier.J.ApplicationSetup.setProperty("TaggedDocument.abstracts.tags.casesensitive", "false")


@pt.java.required
def _FileDocumentSetup(   
        meta : Dict[str,int],  #mapping from meta-key to length
        meta_tags : Dict[str,str] #mapping from meta-key to tag
    ):
    """
    Property setup for FileDocument etc to generate abstracts
    """

    meta_name_for_abstract = None
    for k, v in meta_tags.items():
        if v == 'ELSE':
            meta_name_for_abstract = k
    if meta_name_for_abstract is None:
        return
    
    if meta_name_for_abstract not in meta:
        raise ValueError("You need to specify a meta length for " + meta_name_for_abstract)

    abstract_length = meta[meta_name_for_abstract]

    pt.terrier.J.ApplicationSetup.setProperty("FileDocument.abstract", meta_name_for_abstract)
    pt.terrier.J.ApplicationSetup.setProperty("FileDocument.abstract.length", str(abstract_length))



@pt.java.required
def createAsList(files_path : Union[str, List[str]]):
    """
    Helper method to be used by child indexers to add files to Java List
    Returns:
        Created Java List
    """
    if isinstance(files_path, str):
        asList = pt.java.J.Arrays.asList(files_path)
    elif isinstance(files_path, list):
        asList = pt.java.J.Arrays.asList(*files_path)
    else:
        raise ValueError(f"{files_path}: {type(files_path)} must be a List[str] or str")
    return asList

# Using enum class create enumerations
class IndexingType(Enum):
    """
        This enum is used to determine the type of index built by Terrier. The default is CLASSIC. For more information,
        see the relevant Terrier `indexer <https://terrier-core.readthedocs.io/en/latest/indexer_details.html>`_
        and `realtime <https://terrier-core.readthedocs.io/en/latest/realtime_indices.html>`_ documentation.
    """
    CLASSIC = 1 #: A classical indexing regime, which also creates a direct index structure, useful for query expansion
    SINGLEPASS = 2 #: A single-pass indexing regime, which builds an inverted index directly. No direct index structure is created. Typically is faster than classical indexing.
    MEMORY = 3 #: An in-memory index. No persistent index is created.


@pt.java.required
class TerrierIndexer:
    """
    This is the super class for all of the Terrier-based indexers exposed by PyTerrier. It hosts common configuration
    for all index types.
    
    """
    schematic = {'label': 'TerrierIndexer'}

    default_properties = {
            "TrecDocTags.doctag": "DOC",
            "TrecDocTags.idtag": "DOCNO",
            "TrecDocTags.skip": "DOCHDR",
            "TrecDocTags.casesensitive": "false",
            "trec.collection.class": "TRECCollection",
    }

    def __init__(self,
            index_path : str,
            *,
            blocks : bool = False,
            overwrite: bool = False,
            verbose : bool = False,
            meta_reverse : List[str] = ["docno"],
            stemmer : Union[None, str, TerrierStemmer] = TerrierStemmer.porter,
            stopwords : Union[None, TerrierStopwords, List[str]] = TerrierStopwords.terrier,
            tokeniser : Union[str,TerrierTokeniser] = TerrierTokeniser.english,
            type=IndexingType.CLASSIC,
            properties : Dict[str,str] = {}
    ):
        """
        Constructor called by all indexer subclasses. All arguments listed below are available in 
        IterDictIndexer, DFIndexer, TRECCollectionIndexer and FilesIndsexer. 

        :param index_path: Directory to store index. Ignored for IndexingType.MEMORY.
        :param blocks: Create indexer with blocks if true, else without blocks. Default is False.
        :param overwrite: If index already present at `index_path`, True would overwrite it, False throws an Exception. Default is False.
        :param verbose: Provide progess bars if possible. Default is False.
        :param stemmer: the stemmer to apply. Default is ``TerrierStemmer.porter``.
        :param stopwords: the stopwords list to apply. Default is ``TerrierStemmer.terrier``.
        :param tokeniser: the stemmer to apply. Default is ``TerrierTokeniser.english``.
        :param type: the specific indexing procedure to use. Default is ``IndexingType.CLASSIC``.
        :param properties: Terrier properties that you wish to overrride.
        """
        if type is IndexingType.MEMORY:
            self.path = None
        else:
            self.path = os.path.join(index_path, "data.properties")
            if not os.path.isdir(index_path):
                os.makedirs(index_path)
        self.index_called = False
        self.index_dir = index_path
        self.blocks = blocks
        self.type = type
        self.stemmer = TerrierStemmer._to_obj(stemmer)
        self.stopwords, self.stopword_list = TerrierStopwords._to_obj(stopwords)
        self.tokeniser = TerrierTokeniser._to_obj(tokeniser)
        self.properties = pt.java.J.Properties()
        self.setProperties(**self.default_properties)
        for k,v in properties.items():
            self.properties[k] = v        
        self.overwrite = overwrite
        self.verbose = verbose
        self.meta_reverse = meta_reverse
        self.cleanup_hooks: List[Callable[[TerrierIndexer, Any], None]] = []

    def setProperty(self, k, v):
        """
        Set the named property to the specified value.

        Args:
            k(str): name of the Terrier property
            v(str): value of the Terrier property

        Usage::
            indexer.setProperty("termpipelines", "")
        """
        self.properties.put(str(k), str(v))

    def setProperties(self, **kwargs):
        """
        Set the properties to the given ones.

        Args:
            **kwargs: Properties to set to.

        Usage:
            >>> setProperties(**{property1:value1, property2:value2})
        """
        for key, value in kwargs.items():
            self.properties.put(str(key), str(value))

    def checkIndexExists(self):
        """
        Check if index exists at the `path` given when object was created
        """
        if self.path is None:
            return
        if os.path.isfile(self.path):
            if not self.overwrite:
                raise ValueError("Index already exists at " + self.path)
        if self.index_called:
            raise Exception("Index method can be called only once")

    def createIndexer(self):
        """
        Checks `self.type` and
        - if false, check `blocks` and create a BlockIndexer if true, else create BasicIndexer
        - if true, check `blocks` and create a BlockSinglePassIndexer if true, else create BasicSinglePassIndexer
        Returns:
            Created  Java object extending org.terrier.structures.indexing.Indexer
        """
        
        Indexer, _ = self.indexerAndMergerClasses()
        if Indexer is pt.terrier.J.BasicMemoryIndexer:
            index = Indexer()
        else:
            index = Indexer(self.index_dir, "data")
        assert index is not None
        return index

    def indexerAndMergerClasses(self):
        """
        Check `single_pass` and
        - if false, check `blocks` and create a BlockIndexer if true, else create BasicIndexer
        - if true, check `blocks` and create a BlockSinglePassIndexer if true, else create BasicSinglePassIndexer
        Returns:
            type objects for indexer and merger for the given configuration
        """

        # configure the meta index
        self.properties['indexer.meta.forward.keys'] = ','.join(self.meta.keys())
        self.properties['indexer.meta.forward.keylens'] = ','.join([str(le) for le in self.meta.values()])
        self.properties['indexer.meta.reverse.keys'] = ','.join(self.meta_reverse)

        # configure the term pipeline
        if 'termpipelines' in self.properties:
            # use existing configuration if present
            warn(
                "Setting of termpipelines property directly is deprecated", stacklevel=4, category=DeprecationWarning)
        else:
            
            termpipeline = []
            TerrierStopwords._indexing_config(self.stopwords, self.stopword_list, termpipeline, self.properties, self.cleanup_hooks)

            stemmer_clz = TerrierStemmer._to_class(self.stemmer)
            if stemmer_clz is not None:
                termpipeline.append(stemmer_clz)
            
            self.properties['termpipelines'] = ','.join(termpipeline)

        if "tokeniser" in self.properties:
            warn(
                "Setting of tokeniser property directly is deprecated", stacklevel=4, category=DeprecationWarning)
        else:
            self.properties['tokeniser'] = TerrierTokeniser._to_class(self.tokeniser)

        # inform terrier of all properties
        pt.terrier.J.ApplicationSetup.getProperties().putAll(self.properties)

        # now create the indexers 
        if self.type is IndexingType.SINGLEPASS:
            if self.blocks:
                Indexer = pt.terrier.J.BlockSinglePassIndexer
                Merger = pt.terrier.J.BlockStructureMerger
            else:
                Indexer = pt.terrier.J.BasicSinglePassIndexer
                Merger = pt.terrier.J.StructureMerger
        elif self.type is IndexingType.CLASSIC:
            if self.blocks:
                Indexer = pt.terrier.J.BlockIndexer
                Merger = pt.terrier.J.BlockStructureMerger
            else:
                Indexer = pt.terrier.J.BasicIndexer
                Merger = pt.terrier.J.StructureMerger
        elif self.type is IndexingType.MEMORY:
            if self.blocks:
                raise Exception("Memory indexing with positions not yet implemented")
            else:
                Indexer = pt.terrier.J.BasicMemoryIndexer
                Merger = None
        else:
            raise Exception("Unknown indexer type")
        assert Indexer is not None
        return Indexer, Merger

    def getIndexStats(self):
        """
        Prints the index statistics

        Note:
            Does not work with notebooks at the moment
        """
        pt.terrier.J.CLITool.main(["indexstats", "-I" + self.path])

    def getIndexUtil(self, util):
        """
        Utilities for displaying the content of an index

        Note:
            Does not work with notebooks at the moment

        Args:
            util: which util to print

        Possible Utils:
            util(string): possible values:
            printbitentry
            printlex
            printlist
            printlistentry
            printmeta
            printposting
            printpostingfile
            printterm
        """
        if not util.startswith("-"):
            util = "-" + util
        pt.terrier.J.CLITool.main(["indexutil", "-I" + self.path, util])


class DFIndexUtils:

    @staticmethod
    def get_column_lengths(df):
        meta2len = dict([(v, df[v].apply(lambda r: len(str(r)) if r is not None else 0).max())for v in df.columns.values])
        # nan values can arise if df is empty. Here we take a metalength of 1 instead.
        meta2len = {k : 1 if math.isnan(le) else le for k, le in meta2len.items()}
        return meta2len

    @staticmethod
    @pt.java.required
    def create_javaDocIterator(text, *args, **kwargs):
        HashMap = pt.java.J.HashMap
        TaggedDocument = pt.terrier.J.TaggedDocument
        StringReader = pt.java.J.StringReader
        tokeniser = pt.terrier.J.Tokeniser.getTokeniser()

        all_metadata = {}
        for i, arg in enumerate(args):
            if isinstance(arg, pd.Series):
                all_metadata[arg.name] = arg
                assert len(arg) == len(text), "Length of metadata arguments needs to be equal to length of text argument"
            elif isinstance(arg, pd.DataFrame):
                for name, column in arg.items():
                    all_metadata[name] = column
                    assert len(column) == len(text), "Length of metadata arguments needs to be equal to length of text argument"
            else:
                raise ValueError(f"Non-keyword args need to be of type pandas.Series or pandas.DataFrame, argument {i} was {type(arg)}")
        for key, value in kwargs.items():
            if isinstance(value, (pd.Series, list, tuple)):
                all_metadata[key] = value
                assert len(value) == len(value), "Length of metadata arguments needs to be equal to length of text argument"
            elif isinstance(value, pd.DataFrame):
                for name, column in arg.items():
                    all_metadata[name] = column
                    assert len(column) == len(text), "Length of metadata arguments needs to be equal to length of text argument"
            else:
                raise ValueError("Keyword kwargs need to be of type pandas.Series, list or tuple")

        if "docno" not in all_metadata:
            raise ValueError('No docno column specified, while PyTerrier assumes a docno should exist. Found meta columns were %s' % str(list(all_metadata.keys())))
        # this method creates the documents as and when needed.
        def convertDoc(text_row, meta_column):
            if text_row is None:
                text_row = ""
            hashmap = HashMap()
            for column, value in meta_column[1].items():
                if value is None:
                    value = ""
                hashmap.put(column, value)
            return TaggedDocument(StringReader(text_row), hashmap, tokeniser)
            
        df = pd.DataFrame.from_dict(all_metadata, orient="columns")
        lengths = DFIndexUtils.get_column_lengths(df)
        return (
            PythonListIterator(
                text.values,
                df.iterrows(),
                convertDoc,
                len(text.values),
                ),
            lengths)

@deprecated(version='0.11.0', reason="use pt.terrier.IterDictIndexer().index(dataframe.to_dict(orient='records')) instead")
class DFIndexer(TerrierIndexer):
    """
    Use this Indexer if you wish to index a pandas.Dataframe

    """
    @pt.java.required
    def index(self, text, *args, **kwargs):
        """
        Index the specified

        Args:
            text(pd.Series): A pandas.Series(a column) where each row is the body of text for each document
            *args: Either a pandas.Dataframe or pandas.Series.
                If a Dataframe: All columns(including text) will be passed as metadata
                If a Series: The Series name will be the name of the metadata field and the body will be the metadata content
            **kwargs: Either a list, a tuple or a pandas.Series
                The name of the keyword argument will be the name of the metadata field and the keyword argument contents will be the metadata content
        """
        self.checkIndexExists()
        # we need to prevent collectionIterator from being GCd, so assign to a variable that outlives the indexer
        collectionIterator, meta_lengths = DFIndexUtils.create_javaDocIterator(text, *args, **kwargs)
        
        # record the metadata key names and the length of the values
        self.meta = meta_lengths
        
        # make a Collection class for Terrier
        javaDocCollection = pt.terrier.J.CollectionFromDocumentIterator(collectionIterator)
        if self.verbose:
            javaDocCollection = TQDMSizeCollection(javaDocCollection, len(text)) 
        index = self.createIndexer()
        index.index(pt.terrier.J.PTUtils.makeCollection(javaDocCollection))
        global lastdoc
        lastdoc = None
        javaDocCollection.close()
        self.index_called = True
        collectionIterator = None

        indexref = None
        if self.type is IndexingType.MEMORY:
            index = index.getIndex()
            indexref = index.getIndexRef()
        else:
            indexref = pt.terrier.J.IndexRef.of(self.index_dir + "/data.properties")
            if len(self.cleanup_hooks) > 0:
                sindex = pt.terrier.J.Index
                sindex.setIndexLoadingProfileAsRetrieval(False)
                index = pt.terrier.IndexFactory.of(indexref)
                for hook in self.cleanup_hooks:
                    hook(self, index)
                sindex.setIndexLoadingProfileAsRetrieval(True)
        return indexref


class _BaseIterDictIndexer(TerrierIndexer, pt.Indexer):
    def __init__(self,
                 index_path: str,
                 *,
                 meta : Dict[str,int] = {'docno' : 20},
                 text_attrs : List[str] = ["text"],
                 meta_reverse : List[str] = ['docno'],
                 pretokenised : bool = False,
                 fields : bool = False,
                 threads : int = 1,
                 **kwargs):
        """
        
        :param index_path: Directory to store index. Ignored for IndexingType.MEMORY.
        :param meta: What metadata for each document to record in the index, and what length to reserve. Metadata values will be truncated to this length. Defaults to `{"docno" : 20}`.
        :param text_attrs: List of columns of the input data that should be indexed. These are concatenated in the document representation. Defaults to `["text"]`.
        :param meta_reverse: What metadata should we be able to resolve back to a docid. Defaults to `["docno"]`.
        :param pretokenised: Whether to index pre-tokenized text, e.g., through a Learned Sparse encoder. If True, will ignore ``text_attrs`` and indstead index the dictionary contained in the ``toks`` column.
        :param fields: Whether a fields-indexer should be used, i.e. whether the frequency in each attribute should be recorded separately in the Terrer index. This allows application of weighting models such as BM25F.
        :param threads: Number of threads to use for indexing. Defaults to 1.
        :param kwargs: Additional keyword arguments passed to TerrierIndexer.
        """
        pt.Indexer.__init__(self)
        TerrierIndexer.__init__(self, index_path, **kwargs)

        assert pt.terrier.check_version("5.11"), "Terrier 5.11 is required"

        self.threads = threads
        self.text_attrs = text_attrs
        self.meta = meta
        self.meta_reverse = meta_reverse
        self.pretokenised = pretokenised
        self.fields = fields
        if self.pretokenised:
            if self.fields:
                raise ValueError("pretokenised not supported for fields")
            if self.text_attrs != ['text']: # user provided text_attrs, which is ignored when pretokenised=True
                raise ValueError("pretokenised ignores text_attrs")
            # we disable stemming and stopwords for pretokenised indices
            self.stemmer = None
            self.stopwords = None
            self.text_attrs = ['toks'] # pretokenized always uses toks column

    def _setup(self):
        """
        Configure the indexing properties based on the current indexer configuration.
        """
        self.checkIndexExists()
        if not isinstance(self.meta, dict):
            # user did not specify the lengths of the metadata columns, so set them all to a default value of 512.
            # the ramifications of setting all lengths to a large value is an overhead in memory usage during decompression
            # also increased reverse lookup file if reverse meta lookups are enabled.
            self.meta = {k: '512' for k in self.meta}

        if self.fields:
            self.setProperties(**{
                'metaindex.compressed.crop.long' : 'true',
                'FlatJSONDocument.process' : ','.join(self.text_attrs), # index all these json columns
                'FieldTags.process': ','.join(self.text_attrs), # each of them will be a field for the indexer
                'FieldTags.casesensitive': 'true',
            })
        else:
            self.setProperties(**{
                'metaindex.compressed.crop.long' : 'true',
                'FlatJSONDocument.process' : ','.join(self.text_attrs), # index all these json columns
                'FieldTags.process': '', # but dont make them into fields
                'FieldTags.casesensitive': 'true',
            })

    def _filter_iterable(self, it):
        # Only include necessary columns: those that are indexed, metadata columns, and docno
        # Also, check that the provided iterator is a suitable format

        all_cols = {'docno'} | set(self.text_attrs) | set(self.meta.keys())

        first_docs, it = more_itertools.spy(it) # peek at the first document and validate it
        if len(first_docs) > 0: # handle empty input
            self._validate_doc_dict(first_docs[0])

        # important: return an iterator (i.e. using a generator expression) here, rather than make this 
        # function a generator, to be sure that the validation above happens when  _filter_iterable is 
        # called, rather than on the first invocation of next()
        return ({f: doc[f] for f in all_cols} for doc in it)

    def _is_dict(self, obj):
        return hasattr(obj, '__getitem__') and hasattr(obj, 'items')

    def _validate_doc_dict(self, obj):
        """
        Raise errors/warnings for common indexing mistakes
        """
        if not self._is_dict(obj):
            raise ValueError("Was passed %s while expected dict-like object" % (str(type(obj))))
        if self.meta is not None:
            for k in self.meta:
                if k not in obj:
                    raise ValueError(f"Indexing meta key {k} not found in first document (keys {list(obj.keys())})")
                if len(obj[k]) > int(self.meta[k]):
                    msg = f"Indexing meta key {k} length requested {self.meta[k]} but exceeded in first document (actual length {len(obj[k])}). " + \
                          f"Increase the length in the meta dict for the indexer, e.g., pt.IterDictIndexer(..., meta={ {k: len(obj[k])} })."
                    if k == 'docno':
                        # docnos that are truncated will cause major issues; raise an error
                        raise ValueError(msg)
                    else:
                        # Other fields may not matter as much; just show a warning
                        warn(msg)
    
    def index_inputs(self) -> List[List[str]]:
        return [list(set(list(self.meta.keys()) + self.text_attrs ))]


class _IterDictIndexer_nofifo(_BaseIterDictIndexer):
    """
    Use this Indexer if you wish to index an iter of dicts (possibly with multiple fields).
    This version is used for Windows -- which doesn't support the faster fifo implementation.
    """
    @pt.java.required
    def index(self, it, fields=None):
        """
        Index the specified iter of dicts with the (optional) specified fields

        :param it: an iter of document dicts to be indexed
        """

        if fields is not None:
            if len(fields) > 1:
                raise ValueError("Use fields and text_attrs constructor kwargs")
            raise ValueError("Specify the text attribute to index in the constructor.")

        self._setup()
        assert self.threads == 1, 'IterDictIndexer does not support multiple threads on Windows'

        indexer = self.createIndexer()
        if self.pretokenised:
            assert not self.blocks, "pretokenised isnt compatible with blocks"

            # we generate DocumentPostingList from a dictionary of pretokenised text, e.g.
            # [
            #     {'docno' : 'd1', 'toks' : {'a' : 1, 'aa' : 2}}
            # ]
            
            iter_docs = DocListIterator(self._filter_iterable(it))
            self.index_called = True
            indexer.indexDocuments(iter_docs)
            iter_docs = None

        else:

            # we need to prevent collectionIterator from being GCd
            collectionIterator = FlatJSONDocumentIterator(self._filter_iterable(it))
            javaDocCollection = pt.terrier.J.CollectionFromDocumentIterator(collectionIterator)
            indexer.index(javaDocCollection)
            global lastdoc
            lastdoc = None
            self.index_called = True
            collectionIterator = None

        indexref = None
        if self.type is IndexingType.MEMORY:
            index = indexer.getIndex()
            indexref = index.getIndexRef()
        else:
            indexref = pt.terrier.J.IndexRef.of(self.index_dir + "/data.properties")
            if len(self.cleanup_hooks) > 0:
                sindex = pt.terrier.J.Index
                sindex.setIndexLoadingProfileAsRetrieval(False)
                index = pt.terrier.IndexFactory.of(indexref)
                for hook in self.cleanup_hooks:
                    hook(self, index)
                sindex.setIndexLoadingProfileAsRetrieval(True)

        return indexref

class _IterDictIndexer_fifo(_BaseIterDictIndexer):
    """
    Use this Indexer if you wish to index an iter of dicts (possibly with multiple fields).
    This version is optimized by using multiple threads and POSIX fifos to transfer data,
    which ends up being much faster.
    """
    @pt.java.required
    def index(self, it, fields=None):
        """
        Index the specified iter of dicts with the (optional) specified fields

        :param it: an iter of document dicts to be indexed
        """
        CollectionFromDocumentIterator = pt.terrier.J.CollectionFromDocumentIterator
        JsonlDocumentIterator = pt.terrier.J.JsonlDocumentIterator
        if self.pretokenised:
            JsonlTokenisedIterator = pt.terrier.J.JsonlPretokenisedIterator
        ParallelIndexer = pt.terrier.J.ParallelIndexer

        if fields is not None:
            if len(fields) > 1:
                raise ValueError("Use fields and text_attrs constructor kwargs")
            raise ValueError("Specify the text attribute to index in the constructor.")
        
        self._setup()

        os.makedirs(self.index_dir, exist_ok=True) # ParallelIndexer expects the directory to exist

        Indexer, Merger = self.indexerAndMergerClasses()

        assert self.threads > 0, "threads must be positive"
        if Indexer is pt.terrier.J.BasicMemoryIndexer:
            assert self.threads == 1, 'IterDictIndexer does not support multiple threads for IndexingType.MEMORY'
        if self.threads > 1:
            warn(
                'Using multiple threads results in a non-deterministic ordering of document in the index. For deterministic behavior, use threads=1')

        # Document iterator
        fifos = []
        j_collections = []
        with tempfile.TemporaryDirectory() as d:
            # Make a POSIX FIFO with associated java collection for each thread to use
            for i in range(self.threads):
                fifo = f'{d}/docs-{i}.jsonl'
                os.mkfifo(fifo)
                if (self.pretokenised):
                    j_collections.append(JsonlTokenisedIterator(fifo))
                else:
                    j_collections.append(CollectionFromDocumentIterator(JsonlDocumentIterator(fifo)))
                fifos.append(fifo)

            # Start dishing out the docs to the fifos
            threading.Thread(target=self._write_fifos, args=(self._filter_iterable(it), fifos), daemon=True).start()

            # Different process for memory indexer (still taking advantage of faster fifos)
            if Indexer is pt.terrier.J.BasicMemoryIndexer:
                indexer = Indexer()
                if self.pretokenised:
                    indexer.indexDocuments(j_collections)
                else:
                    indexer.index(j_collections)
                for hook in self.cleanup_hooks:
                    hook(self, indexer.getIndex())
                return indexer.getIndex().getIndexRef()

            # Start the indexing threads
            if self.pretokenised:
                ParallelIndexer.buildParallelTokenised(j_collections, self.index_dir, Indexer, Merger)
            else:
                ParallelIndexer.buildParallel(j_collections, self.index_dir, Indexer, Merger)
            
        indexref = None
        indexref = pt.terrier.J.IndexRef.of(self.index_dir + "/data.properties")
        
        if len(self.cleanup_hooks) > 0:
            sindex = pt.terrier.J.Index
            sindex.setIndexLoadingProfileAsRetrieval(False)
            index = pt.terrier.IndexFactory.of(indexref)
            sindex.setIndexLoadingProfileAsRetrieval(True)

            for hook in self.cleanup_hooks:
                hook(self, index)

        return indexref

    def _write_fifos(self, it, fifos):
        with contextlib.ExitStack() as stack:
            fifos = [stack.enter_context(open(f, 'wt')) for f in fifos]
            ready = None
            for doc in it:
                if not ready: # either first iteration or deque is empty
                    if len(fifos) > 1:
                        # Not all the fifos may be ready yet for the next document. Rather than
                        # waiting for the next one to finish up, go ahead and can check wich are ready
                        # with the select syscall. This will block until at least one is ready. This
                        # optimization can actually have a pretty big impact-- on CORD19, indexing
                        # with 8 threads was 30% faster with this.
                        _, ready, _ = select.select([], fifos, [])
                        ready = deque(ready)
                    else:
                        # single threaded mode
                        ready = deque(fifos)
                fifo = ready.popleft()
                if 'toks' in doc:
                    doc['toks'] = {k: int(v) for k, v in doc['toks'].items()} # cast all values as ints
                json.dump(doc, fifo)
                fifo.write('\n')


# Windows doesn't support fifos -- so we have 2 versions.
# Choose which one to expose based on whether os.mkfifo exists.
IterDictIndexer: Type[Union[_IterDictIndexer_fifo, _IterDictIndexer_nofifo]]
if hasattr(os, 'mkfifo'):
    #IterDictIndexer = _IterDictIndexer_nofifo
    IterDictIndexer = _IterDictIndexer_fifo
else:
    IterDictIndexer = _IterDictIndexer_nofifo
 # trick sphinx into not using "alias of"
IterDictIndexer.__name__ = 'IterDictIndexer'

class TRECCollectionIndexer(TerrierIndexer):


    """
        Use this Indexer if you wish to index a TREC formatted collection
    """

    def __init__(self, 
            index_path : str, 
            collection : str = "trec", 
            verbose : bool = False,
            meta : Dict[str,int] = {"docno" : 20},
            meta_reverse : List[str] = ["docno"],
            meta_tags : Dict[str,str] = {},
            **kwargs
            ):
        """
        Init method

        :param index_path: Directory to store index. Ignored for IndexingType.MEMORY.
        :param blocks: Create indexer with blocks if true, else without blocks. Default is False.
        :param overwrite: If index already present at `index_path`, True would overwrite it, False throws an Exception. Default is False.
        :param type: the specific indexing procedure to use. Default is IndexingType.CLASSIC.
        :param collection: name, or Class instance, or one of "trec", "trecweb", "warc"). Default is "trec".
        :param meta: What metadata for each document to record in the index, and what length to reserve. Metadata fields will be truncated to this length. Defaults to `{"docno" : 20}`.
        :param meta_reverse: What metadata shoudl we be able to resolve back to a docid. Defaults to `["docno"]`.
        :param meta_tags: For collections formed using tagged data (e.g. HTML), which tags correspond to which metadata. This is useful for recording the text of documents for use in neural rankers - see :ref:`pt.text`.

        """
        super().__init__(index_path, **kwargs)
        if isinstance(collection, str):
            if collection in type_to_class:
                collection = type_to_class[collection]
        self.collection = collection
        self.verbose = verbose
        self.meta = meta
        self.meta_reverse = meta_reverse
        self.meta_tags = meta_tags
    
    @pt.java.required
    def index(self, files_path : Union[str,List[str]]):
        """
        Index the specified TREC formatted files

        Args:
            files_path: can be a String of the path or a list of Strings of the paths for multiple files
        """
        self.checkIndexExists()
        index = self.createIndexer()
        if not isinstance(files_path, list):
            raise ValueError('files_path must be a list')

        _TaggedDocumentSetup(self.meta, self.meta_tags)

        colObj = createCollection(files_path, self.collection)
        if self.verbose and isinstance(colObj, pt.terrier.J.MultiDocumentFileCollection):
            colObj = pt.java.cast("org.terrier.indexing.MultiDocumentFileCollection", colObj)
            colObj = TQDMCollection(colObj)
        # remove once 5.7 is now the minimum version
        if pt.terrier.check_version("5.7"):
            index.index(colObj)
        else:
            index.index(pt.terrier.J.PTUtils.makeCollection(colObj))
        global lastdoc
        lastdoc = None
        colObj.close()
        self.index_called = True
        
        for hook in self.cleanup_hooks:
            hook(self, index.getIndex())

        if self.type is IndexingType.MEMORY:
            return index.getIndex().getIndexRef()
        return pt.terrier.J.IndexRef.of(self.index_dir + "/data.properties")

class FilesIndexer(TerrierIndexer):
    '''
        Use this Indexer if you wish to index a pdf, docx, txt etc files

        Args:
            index_path (str): Directory to store index. Ignored for IndexingType.MEMORY.
            blocks (bool): Create indexer with blocks if true, else without blocks. Default is False.
            type (IndexingType): the specific indexing procedure to use. Default is IndexingType.CLASSIC.
            meta(Dict[str,int]): What metadata for each document to record in the index, and what length to reserve. Metadata fields will be truncated to this length. Defaults to `{"docno" : 20, "filename" : 512}`.
            meta_reverse(List[str]): What metadata shoudl we be able to resolve back to a docid. Defaults to `["docno"]`,
            meta_tags(Dict[str,str]): For collections formed using tagged data (e.g. HTML), which tags correspond to which metadata. Defaults to empty. This is useful for recording the text of documents for use in neural rankers - see :ref:`pt.text`.

    '''

    def __init__(self, index_path, *, meta={"docno" : 20, "filename" : 512}, meta_reverse=["docno"], meta_tags={}, **kwargs):
        super().__init__(index_path, **kwargs)
        self.meta = meta
        self.meta_reverse = meta_reverse
        self.meta_tags = meta_tags

    @pt.java.required
    def index(self, files_path : Union[str,List[str]]):
        """
        Index the specified files.

        :param files_path: can be a String of the path or a list of Strings of the paths for multiple files
        """
        self.checkIndexExists()
        index = self.createIndexer()
        asList = createAsList(files_path)
        _TaggedDocumentSetup(self.meta, self.meta_tags)
        _FileDocumentSetup(self.meta, self.meta_tags)
        
        simpleColl = pt.terrier.J.SimpleFileCollection(asList, False)
        # remove once 5.7 is now the minimum version
        index.index(simpleColl if pt.terrier.check_version("5.7") else [simpleColl])
        global lastdoc
        lastdoc = None
        self.index_called = True

        indexref = None
        if self.type is IndexingType.MEMORY:
            index = index.getIndex()
            indexref = index.getIndexRef()
        else:
            indexref = pt.terrier.J.IndexRef.of(self.index_dir + "/data.properties")
            if len(self.cleanup_hooks) > 0:
                sindex = pt.terrier.J.Index
                sindex.setIndexLoadingProfileAsRetrieval(False)
                index = pt.terrier.IndexFactory.of(indexref)
                for hook in self.cleanup_hooks:
                    hook(self, index)
                sindex.setIndexLoadingProfileAsRetrieval(True)
        return indexref



from xml.dom import minidom
import urllib.request
from packaging.version import Version
from pathlib import Path
from enum import Enum
from warnings import warn
import requests
import os
import re
import pyterrier as pt

MAVEN_BASE_URL = "https://repo1.maven.org/maven2/"
JITPACK_BASE_URL = "https://jitpack.io/"
USER_AGENT = "curl/7.81.0"

class OnlineMode(Enum):
    ONLINE = 'ONLINE'
    OFFLINE = 'OFFLINE'
    UNSET = 'UNSET'


_online_mode = OnlineMode.UNSET


def online_mode() -> OnlineMode:
    """
    Returns whether mavenresolver is operating in "online" mode, "offline" mode, or if the mode has not been set yet.

    In "online" mode, packages will always be requested from maven. Network failures will raise errors in this mode.

    In "offline" mode, maven will not be contacted, and the most recent versions of packages that are currently cached
    on the system will be used when requesting package versions, etc. If no versions are available, an error will be
    raised.

    If not set, mavenresolver will behave in "online" mode until it receives a network error, at which point it will
    move automatically to "offline" mode.
    """
    return _online_mode


def offline(is_offline: bool = True):
    """
    Enables offline mode. 
    """
    global _online_mode
    _online_mode = {
        True: OnlineMode.OFFLINE,
        False: OnlineMode.ONLINE
    }[is_offline]


def online(is_online: bool = True):
    global _online_mode
    _online_mode = {
        True: OnlineMode.ONLINE,
        False: OnlineMode.OFFLINE
    }[is_online]


# obtain a file from maven
def get_package_jar(orgName, packageName, version, file_path=None, artifact="jar", force_download=False):
    if file_path is None:
        file_path = pt.io.pyterrier_home()
    orgName = orgName.replace(".", "/")
    suffix = ""
    ext = "jar"
    if artifact == "jar-with-dependencies" or artifact == "fatjar":
        suffix = "-" + artifact
        ext = "jar"
    if artifact == "pom":
        ext = "pom"
    filename = packageName + "-" + version + suffix + "." + ext

    filelocation = orgName + "/" + packageName + "/" + version + "/" + filename

    mode = online_mode()

    if force_download and mode == OnlineMode.OFFLINE:
        force_download = False # OFFLINE overrides force_download
    
    target_file = os.path.join(file_path, filename)
    file_exists = os.path.isfile(target_file)
    if file_exists:
        if not force_download:
            return target_file
        else:
            # ensure that we put the file in a different name
            os.rename(target_file, f'{target_file}.bak')

    # check local Maven repo, and use that if it exists
    from os.path import expanduser
    userhome = expanduser("~")
    mavenRepoHome = os.path.join(userhome, ".m2", "repository")
    mvnLocalLocation = os.path.join(mavenRepoHome, filelocation)
    if os.path.isfile(mvnLocalLocation):
        return mvnLocalLocation

    if mode == OnlineMode.OFFLINE:
        raise ValueError(f'Offline mode, and {filename} not found')

    if force_download:
        print("Downloading "+ packageName + " " + version + " " + artifact  + " to " + file_path + "...")
    else:
        print(packageName + " " + version + " " + artifact  + " not found, downloading to " + file_path + "...")
    
    if "com/github" in orgName:
        mvnUrl = JITPACK_BASE_URL + filelocation
    else:
        mvnUrl = MAVEN_BASE_URL + filelocation

    try:
        pt.io.download(mvnUrl, target_file, verbose=True, headers={"User-Agent": USER_AGENT})
    except requests.exceptions.ConnectionError as he:
        if mode == OnlineMode.UNSET:
            offline() # now we're in offline mode
        if file_exists:
            os.rename(f'{target_file}.bak', target_file) # move back
            if mode == OnlineMode.UNSET:
                warn(f'Attempted to download {mvnUrl}, but was offline. Using cached version: {target_file}')
                return target_file
        raise ValueError("Could not fetch " + mvnUrl) from he

    if file_exists:
        os.remove(f'{target_file}.bak') # clean up temp file
    print("Done")

    return os.path.join(file_path, filename)

# returns the latest version
def latest_version_num(orgName, packageName):
    orgName = orgName.replace(".", "/")
    if "com/github" in orgName:
        # its jitpack
        return "-SNAPSHOT"

    mode = online_mode()

    if mode == OnlineMode.OFFLINE:
        version = latest_local_version_num(packageName)
        if version is None:
            raise ValueError(f'Could not find latest version of {packageName}')
        return version

    url_str = MAVEN_BASE_URL + orgName + "/" + packageName + "/maven-metadata.xml"
    try:
        with urllib.request.urlopen(urllib.request.Request( url_str, headers={"User-Agent": USER_AGENT})) as url:
            xml_str = url.read()
    except urllib.error.URLError as ue:
        if mode == OnlineMode.UNSET:
            offline()
            # no internet connection, use the latest version found locally.
            version = latest_local_version_num(packageName)
            if version is None:
                raise # version not found, re-raise the URLError error
            else:
                warn(f'Attempted to get latest version of {packageName} from maven, but was offline ({url_str} {ue.code} {ue.reason}). Using latest cached version: {version}')
                return version
        else: # mode == OnlineMode.ONLINE
            raise
    xmldoc = minidom.parseString(xml_str)
    obs_values = xmldoc.getElementsByTagName("latest")
    version = obs_values[0].firstChild.nodeValue
    return version


def latest_local_version_num(packageName):
    versions = []
    for jar_path in Path(pt.io.pyterrier_home()).glob(f'{packageName}-*.jar'):
        match = re.search(rf'{packageName}-([0-9]+(\.[0-9]+)+).*.jar', jar_path.name)
        if match:
            versions.append(Version(match.group(1)))
    if len(versions) > 0:
        return str(max(versions))
    return None # no local version found



import os
import warnings
import traceback
import types
import pandas as pd
from typing import Iterator, List, Union, Tuple, Protocol, runtime_checkable, Optional, Any
import pyterrier as pt

LAMBDA = lambda:0  # noqa: E731 LAMBDA is used for the is_lambda method below, so the type is important
def is_lambda(v):
    return isinstance(v, type(LAMBDA)) and v.__name__ == LAMBDA.__name__

def is_function(v):
    return isinstance(v, types.FunctionType)

def is_transformer(v):
    if isinstance(v, Transformer):
        return True
    return False

def get_transformer(v, stacklevel=1):
    """ 
        Used to coerce functions, lambdas etc into transformers 
    """

    if is_transformer(v):
        return v
    if is_lambda(v):
        raise ValueError('Coercion of a lambda into a transformer is no longer supported; use a pt.apply instead')
    if is_function(v):
        raise ValueError('Coercion of a function (called "%s") into a transformer is no longer supported; use a pt.apply instead' % v.__name__)
    if isinstance(v, pd.DataFrame):
        raise ValueError('Coercion of a dataframe into a transformer is no longer supported; use a pt.Transformer.from_df() instead')
    raise ValueError("Passed parameter %s of type %s cannot be coerced into a transformer" % (str(v), type(v)))


class Transformer:
    """
        Base class for all transformers. Implements the various operators ``>>`` ``+`` ``*`` ``|`` ``&`` 
        as well as :meth:`search` for executing a single query and :meth:`compile` for rewriting complex pipelines into more simples ones.

        Its expected that either :meth:`transform` or :meth:`transform_iter()` be implemented by any class extending this - this rule
        does not apply for indexers, which instead implement ``.index()``. pt.apply helper functions can be used to
        easily construct Transformers around a single function.
    """
    name = "Transformer"

    def __new__(cls, *args, **kwargs):
        if cls.transform == Transformer.transform and cls.transform_iter == Transformer.transform_iter:
            raise NotImplementedError("You need to implement either .transform() or .transform_iter() in %s" % str(cls))
        
        instance = super().__new__(cls)
        if cls.transform == Transformer.transform and cls.transform_iter != Transformer.transform_iter:
            # User implemented transform_iter on this transformer but not transform.
            # transform_iter is not inspectable, so place a transform_outputs method on the instance if not exists
            if not hasattr(instance, 'transform_outputs'):
                def _transform_outputs(self, inp_cols):
                    raise pt.inspect.InspectError("This transformer is not inspectable - you can override/implement the transform_outputs method to make it inspectable")
                instance.transform_outputs = types.MethodType(_transform_outputs, instance)
        return instance

    @staticmethod
    def identity() -> 'Transformer':
        """
        Instantiates a transformer that returns exactly its input. 
        
        This can be useful for adding the candidate ranking score
        as a feature in for learning-to-rank::

            bm25 = pt.terrier.Retriever(index, wmodel="BM25")
            two_feat_pipe = bm25 >> pt.Transformer.identity() ** pt.terrier.Retriever(index, wmodel="PL2")
        
        This will return a pipeline that produces a score column (BM25), but also has a features column containing
        BM25 and PL2 scores.
        
        """
        return IdentityTransformer()

    @staticmethod
    def from_df(input : pd.DataFrame, uniform : bool = False) -> 'Transformer':
        """
        Instantiates a transformer from an input dataframe. Some rows from the input dataframe are returned
        in response to a query on the :meth:`transform` method. Depending on the value `uniform`, the dataframe
        passed as an argument to :meth:`transform` can affect this selection.

        :param input: a dataframe to store and return, based on setting of `uniform`.
        :param uniform: If True, input will be returned in its entirety each time, else rows from input that match the qid values from the argument dataframe.
        
        """
        if uniform:
            return UniformTransformer(input)
        return SourceTransformer(input)

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """
            Abstract method that runs the transformer over Pandas ``DataFrame`` objects. This or :meth:`transform_iter`
            must be implemented by all Transformer objects.

            .. note::

                Either :meth:`transform` or :meth:`transform_iter` must be implemented for all transformers.
                If not, a runtime error will be raised when constructing the transformer.

                When :meth:`transform` is not implemented, the default implementation runs :meth:`transform_iter` and
                converts the output to a ``DataFrame``.

            :param inp: The input to the transformer (e.g., queries, documents, results, etc.)
            :return: The output of the transformer (e.g., result of retrieval, re-writing, re-ranking, etc.)
            :rtype: ``pd.DataFrame``
        """
        # We should have no recursive transform <-> transform_iter problem, due to the __new__ check.
        return pd.DataFrame(list(self.transform_iter(inp.to_dict(orient='records'))))

    def transform_iter(self, inp: pt.model.IterDict) -> pt.model.IterDict:
        """
            Abstract method that runs the transformer over iterable input (such as lists or generators),
            where each element is a dictionary record. This or :meth:`transform`
            must be implemented by all Transformer objects.

            This format can sometimes be easier to implement than :meth:`transform`. Furthermore, it avoids constructing
            expensive ``DataFrame`` objects. It is also used in the invocation of ``index()`` on a composed pipeline.

            .. note::

                Either :meth:`transform` or :meth:`transform_iter` must be implemented for all transformers.
                If not, a runtime error will be raised when constructing the transformer.

                When :meth:`transform_iter` is not implemented, the default implementation runs :meth:`transform` and
                converts the output to an iterable.

            :param inp(``Iterable[Dict]``): The input to the transformer (e.g., queries, documents, results, etc.)
            :return: The output of the transformer (e.g., result of retrieval, re-writing, re-ranking, etc.)
            :rtype: ``Iterable[Dict]``
        """
        # We should have no recursive transform <-> transform_iter problem, due to the __new__ check.
        return self.transform(pd.DataFrame(list(inp))).to_dict(orient='records')
    
    def __call__(self, inp: Union[pd.DataFrame, pt.model.IterDict, List[pt.model.IterDictRecord]]) -> Union[pd.DataFrame, pt.model.IterDict, List[pt.model.IterDictRecord]]:
        """
            Runs the transformer for the given input and returns its output as the same type as the input.

            - When ``inp`` is a DataFrame, invokes :meth:`transform` and returns a DataFrame
            - When ``inp`` is a list, invokes :meth:`transform_iter` and returns a list
            - Otherwise, invokes :meth:`transform_iter` and returns a generic iterable (returning whatever type is
              returned from :meth:`transform_iter()`.)

            :param inp: The input to the transformer (e.g., queries, documents, results, etc.)
            :return: The output of the transformer (e.g., result of retrieval, re-writing, re-ranking, etc.) as the same
                type as the input.
            :rtype: ``pd.DataFrame``, ``Iterable[Dict]``, ``List[Dict]``
        """
        if isinstance(inp, pd.DataFrame):
            return self.transform(inp)
        out = self.transform_iter(inp)
        if isinstance(inp, list):
            return list(out)
        return out

    def transform_gen(self, input : pd.DataFrame, batch_size=1, output_topics=False) -> Union[Iterator[pd.DataFrame], Iterator[Tuple[pd.DataFrame, pd.DataFrame]]]:
        """
            Method for executing a transformer pipeline on smaller batches of queries.
            The input dataframe is grouped into batches of batch_size queries, and a generator
            returned, such that :meth:`transform` is only executed for a smaller batch at a time. 

            :param input: a dataframe to process
            :param batch_size: how many input instances to execute in each batch. Defaults to 1.
            
        """
        docno_provided = "docno" in input.columns
        docid_provided = "docid" in input.columns
        
        if docno_provided or docid_provided:
            queries = input[["qid"]].drop_duplicates()
        else:
            queries = input
        batch : List[pd.DataFrame] = []      
        for query in queries.itertuples():
            if len(batch) == batch_size:
                batch_topics = pd.concat(batch)
                batch=[]
                res = self.transform(batch_topics)
                if output_topics:
                    yield res, batch_topics
                else:
                    yield res
            batch.append(input[input["qid"] == query.qid])
        if len(batch) > 0:
            batch_topics = pd.concat(batch)
            res = self.transform(batch_topics)
            if output_topics:
                yield res, batch_topics
            else:
                yield res

    def search(self, query : str, qid : str = "1", sort : bool = True) -> pd.DataFrame:
        """
            Method for executing a transformer (pipeline) for a single query. 

            :param query: String form of the query to run
            :param qid: the query id to associate to this request. defaults to 1.
            :param sort: ensures the results are sorted by descending rank (defaults to True)

            :return: Returns a dataframe with the results for the specified query. This
                is a utility method, and most uses are expected to use the :meth:`transform`
                method passing a dataframe.

            Example::

                bm25 = pt.terrier.Retriever(index, wmodel="BM25")
                res = bm25.search("example query")

                # is equivalent to
                queryDf = pd.DataFrame([["1", "example query"]], columns=["qid", "query"])
                res = bm25.transform(queryDf)
            
            
        """
        import pandas as pd
        queryDf = pd.DataFrame([[qid, query]], columns=["qid", "query"])
        rtr = self.transform(queryDf)
        if "qid" in rtr.columns and "rank" in rtr.columns:
            rtr = rtr.sort_values(["qid", "rank"], ascending=[True,True])
        return rtr

    def compile(self) -> 'Transformer':
        """Returns an optimised transformer, if possible, to improve performance.

        For instance, a pipeline of transformers can be optimised by fusing adjacent transformers.

        :return: A new transformer that is equivalent to this transformer, but optimised.
        """
        return self # by default, nothing to compile

    def parallel(self, N : int, backend : str='joblib') -> 'Transformer':
        """
        Returns a parallelised version of this transformer. The underlying transformer must be "picklable". For more information, see
        :ref:`parallel` documentation.

        :param N: how many processes/machines to parallelise this transformer over.
        :param backend: which multiprocessing backend to use. Only two backends are supported, 'joblib' and 'ray'. Defaults to 'joblib'.
        """
        from warnings import warn
        warn(".parallel() is experimental")
        try:
            from pyterrier_alpha.parallel import PoolParallelTransformer # type: ignore
        except ImportError as ie:
            raise ImportError("pyterrier-alpha[parallel] must be installed for .parallel()") from ie
        return PoolParallelTransformer(self, N, backend)

    # Get and set specific parameter value by parameter's name
    def get_parameter(self, name : str):
        """
            Gets the current value of a particular key of the transformer's configuration state.
            By default, this examines the attributes of the transformer object, using ``hasattr()`` and ``setattr()``.

            :param name: name of parameter
        """
        if hasattr(self, name):
            return getattr(self, name)
        else:
            raise ValueError(("Invalid parameter name %s for transformer %s. " + 
                      "Check the list of available parameters") %(str(name), str(self)))

    def set_parameter(self, name : str, value):
        """
            Adjusts this transformer's configuration state, by setting the value for specific parameter.
            By default, this examines the attributes of the transformer object, using ``hasattr()`` and ``setattr()``.

            :param name: name of parameter
            :param value: current value of parameter

        """
        if hasattr(self, name):
            setattr(self, name, value)
        else:
            raise ValueError(('Invalid parameter name %s for transformer %s. '+
                    'Check the list of available parameters') %(name, str(self)))

    def __rshift__(self, right) -> 'Transformer':
        from ._ops import Compose
        return Compose(self, right)

    def __rrshift__(self, left) -> 'Transformer':
        from ._ops import Compose
        return Compose(left, self)

    def __add__(self, right : 'Transformer') -> 'Transformer':
        from ._ops import Sum
        return Sum(self, right)

    def __pow__(self, right : 'Transformer') -> 'Transformer':
        from ._ops import FeatureUnion
        return FeatureUnion(self, right)

    def __mul__(self, rhs : Union[float,int]) -> 'Transformer':
        assert isinstance(rhs, int) or isinstance(rhs, float)
        from ._ops import ScalarProduct
        return self >> ScalarProduct(rhs)

    def __rmul__(self, lhs : Union[float,int]) -> 'Transformer':
        assert isinstance(lhs, int) or isinstance(lhs, float)
        from ._ops import ScalarProduct
        return self >> ScalarProduct(lhs)

    def __or__(self, right : 'Transformer') -> 'Transformer':
        from ._ops import SetUnion
        return SetUnion(self, right)

    def __and__(self, right : 'Transformer') -> 'Transformer':
        from ._ops import SetIntersection
        return SetIntersection(self, right)

    def __mod__(self, right : int) -> 'Transformer':
        assert isinstance(right, int)
        from ._ops import RankCutoff
        return self >> RankCutoff(right)

    def __xor__(self, right : 'Transformer') -> 'Transformer':
        from ._ops import Concatenate
        return Concatenate(self, right)

    def __invert__(self : 'Transformer') -> 'Transformer':
        raise NotImplementedError("Use pyterrier-caching for more fine-grained caching, e.g. RetrieverCache or ScorerCache")

    def __hash__(self):
        return hash(repr(self))

    def _repr_html_(self):
        if os.environ.get('PYTERRIER_DISABLE_NOTEBOOK_SCHEMATIC', '0') == '1':
            return None
        try:
            return pt.schematic.draw(self, outer_class='repr_html')
        except Exception as e:
            # This handles cases where there's a problem generating the schematic.
            # signal to ipython not to display HTML representation (falls back on __repr__)
            tb = traceback.format_exc()
            warnings.warn(f"Failed to render transformer as HTML: {e}\n\n{tb}", stacklevel=2)
            return None 


class Indexer(Transformer):
    def __new__(cls, *args, **kwargs):
        instance = super().__new__(cls, *args, **kwargs)
        # We have some patching do to in the (somewhat rare) case where the user implements transform/transform_iter
        # on an indexer. Normally these raise errors when called on an indexer, but in this case the user wants the
        # indexer to act as a transformer. So patch the complementary method to call the implemented one.
        if cls.transform != Indexer.transform and cls.transform_iter == Indexer.transform_iter:
            # User implemented transform on this indexer but not transform_iter. Replace transform_iter with the default
            # one, which invokes transform automatically.
            instance.transform_iter = types.MethodType(Transformer.transform_iter, instance)
        elif cls.transform == Indexer.transform and cls.transform_iter != Indexer.transform_iter:
            # User implemented transform_iter on this indexer but not transform. Replace transform with the default
            # one, which invokes transform_iter automatically.
            instance.transform = types.MethodType(Transformer.transform, instance)
        return instance
    
    def index_inputs(self) -> Optional[List[List[str]]]:
        """
            Returns a list of column configurations that index() is expects. 
            This default implementation returns None, and should be 
            overridden by subclasses to allow accurate inspections and schematic visualisations.
        """
        return None

    def index(self, iter : pt.model.IterDict, **kwargs) -> Any:
        """
            Takes an iterable of dictionaries ("iterdict"), and consumes them. The index method may return
            an instance of the index or retriever. This method is typically used to implement indexers that
            consume a corpus (or to consume the output of previous pipeline components that have
            transformer the documents being consumed).

            :param iter: An iterable of dictionaries, each representing a document.
        """
        pass

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError('You called `transform()` on an indexer. Did you mean to call `index()`?')

    def transform_iter(self, inp: pt.model.IterDict) -> pt.model.IterDict:
        raise NotImplementedError('You called `transform_iter()` on an indexer. Did you mean to call `index()`?')

class Estimator(Transformer):
    """
        This is a base class for things that can be fitted.
    """
    def fit(self, topics_or_res_tr, qrels_tr, topics_or_res_va, qrels_va):
        """
            Method for training the transformer.

            :param topics_or_res_tr: training topics (usually with documents)
            :param qrels_tr: training qrels
            :param topics_or_res_va: validation topics (usually with documents)
            :param qrels_va: validation qrels
        """
        pass

class IdentityTransformer(Transformer):
    """
        A transformer that returns exactly the same as its input.
    """
    def __init__(self, *args, **kwargs):
        super(IdentityTransformer, self).__init__(*args, **kwargs)
    
    def transform(self, topics):
        return topics

class SourceTransformer(Transformer):
    """
    A Transformer that can be used when results have been saved in a dataframe.
    It will select results on qid.
    If a column is in the dataframe passed in the constructor, this will override any
    column in the topics dataframe passed to the transform() method.
    """
    def __init__(self, df):
        self.df = df
        assert "qid" in self.df.columns
    
    def transform(self, topics):
        import numpy as np
        assert "qid" in topics.columns
        keeping = topics.columns

        common_columns = np.intersect1d(topics.columns, self.df.columns)

        # we drop columns in topics that exist in the self.df
        drop_columns = common_columns[common_columns != "qid"]
        if len(drop_columns) > 0:
            keeping = topics.columns[~ topics.columns.isin(drop_columns)]
        
        rtr = topics[keeping].merge(self.df, on="qid")
        return rtr

class UniformTransformer(Transformer):
    """
        A transformer that returns the same dataframe every time transform()
        is called. This class is useful for testing. 
    """
    def __init__(self, rtr):
        self.rtr = rtr
    
    def transform(self, topics):
        rtr = self.rtr.copy()
        return rtr

    def __repr__(self):
        return 'UniformTransformer()'
    
    def transform_outputs(self, input_columns: List[str]) -> List[str]:
        return self.rtr.columns.tolist()

@runtime_checkable
class SupportsFuseRankCutoff(Protocol):
    def fuse_rank_cutoff(self, k: int) -> Optional[Transformer]:
        """Fuses this transformer with a following RankCutoff transformer.

        This method should return a new transformer that applies the new rank cutoff value `k`.

        Note that if the transformer currently applies a stricter rank cutoff than the one provided, it should not be
        relaxed. In this case, it is preferred to return `self`.

        If the fusion is not possible, `None` should be returned.

        :param k: The rank cutoff requested
        """


@runtime_checkable
class SupportsFuseFeatureUnion(Protocol):
    def fuse_feature_union(self, other: Transformer, is_left: bool) -> Optional[Transformer]:
        """Fuses this transformer with another one that provides features.

        This method should return a new transformer that is equivalent to performing self ** other, or `None`
        if the fusion is not possible.

        :param other: transformer to the left or right.
        :param is_left: is True if self's features are to the left of other's. Otherwise, self's features are to the right.
        """


@runtime_checkable
class SupportsFuseLeft(Protocol):
    def fuse_left(self, left: Transformer) -> Optional[Transformer]:
        """Fuses this transformer with a transformer that immediately precedes this one in a composed (`>>`) pipeline.

        The new transformer should have the same effect as performing the two transformers in sequence, i.e.,
        `pipeline_unfused` and `pipeline_fused` in the following example should provide the same results for
        any input::

            pipeline_unfused = left >> self
            pipeline_fused = self.fuse_left(left)

        A fused transformer should be more efficient than the unfused version. For instance, a retriever
        followed by a rank cutoff can be fused to perform the rank cutoff during retrieval.

        :param left: transformer to the left.

        :return: A new transformer that is the result of merging this transformer with the left transformer,
            or None if a merge is not possible.
        """


@runtime_checkable
class SupportsFuseRight(Protocol):
    def fuse_right(self, right: Transformer) -> Optional[Transformer]:
        """Fuses this transformer with a transformer that immediately follows this one in a composed (`>>`) pipeline.

        The new transformer should have the same effect as performing the two transformers in sequence, i.e.,
        `pipeline_unfused` and `pipeline_fused` in the following example should provide the same results for
        any input::

            pipeline_unfused = self >> right
            pipeline_fused = self.fuse_right(right)

        A fused transformer should be more efficient than the unfused version. For instance, a retriever
        followed by a rank cutoff can be fused to perform the rank cutoff during retrieval.

        :param right: transformer to the right in a composed pipeline.

        :return: A new transformer that is the result of merging this transformer with the right transformer,
            or None if a merge is not possible.
        """





from docutils import nodes
from docutils.parsers.rst import Directive, directives
from sphinx.util.docutils import SphinxDirective
from sphinx.util import logging
from sphinx import addnodes

logger = logging.getLogger(__name__)


class RelatedDirective(SphinxDirective):
    """
    Directive to mark related items.
    
    Usage:
        .. related:: module.Class.method
           :title: Optional custom title
    """
    required_arguments = 1
    optional_arguments = 0
    option_spec = {
        'title': directives.unchanged,
    }
    has_content = False

    def run(self):
        env = self.env
        target = self.arguments[0]
        
        # Get current document and section info
        docname = env.docname
        
        # Find the current section
        section_id, section_title, page_title = self._get_current_section()
        
        # Store the related reference for later processing
        if not hasattr(env, 'related_items'):
            env.related_items = {}
        
        if target not in env.related_items:
            env.related_items[target] = []
        
        env.related_items[target].append({
            'docname': docname,
            'section_id': section_id,
            'section_title': section_title,
            'page_title': page_title,
        })
        
        # Return empty list - don't output anything where the directive appears
        return []
    
    def _get_current_section(self):
        """Find the current section ID and title."""
        node = self.state.parent
        section_ids = []
        titles = []
        while node:
            if isinstance(node, nodes.section):
                # Get section ID
                section_id = node.get('ids', [None])[0]
                
                # Get section title
                title_node = node.next_node(nodes.title)
                section_title = title_node.astext() if title_node else 'Unknown Section'
                section_ids.append(section_id)
                titles.append(section_title)
            node = node.parent
        
        if section_ids:
            return section_ids[0], titles[0], titles[-1]
        return None, 'Top of Document'


def process_related_items(app, doctree, docname):
    """
    Process related items after doctree is resolved.
    This adds backlinks to the referenced documentation.
    """
    env = app.builder.env
    
    if not hasattr(env, 'related_items'):
        return
    
    # Check if current document is referenced by any related directives
    for obj in doctree.traverse(addnodes.desc):
        # Get the signature node
        sig = obj.next_node(addnodes.desc_signature)
        if not sig:
            continue
        
        # Get the full name of the documented object
        names = sig.get('ids', [])
        
        for name in names:
            if name in env.related_items:
                # Add a "Referenced in" section
                references = env.related_items[name]
                
                # Find the desc_content node (the <dd> element)
                desc_content = obj.next_node(addnodes.desc_content)
                if not desc_content:
                    continue
                
                # Create a seealso admonition
                seealso = nodes.admonition()
                seealso['classes'].append('seealso')
                seealso['classes'].append('related')
                
                # Add title
                title = nodes.title()
                title += nodes.Text('See also')
                seealso += title
                
                for i, ref in enumerate(references):
                    para = nodes.paragraph()
                    
                    # Create reference to the section
                    ref_text = ref['section_title']
                    
                    # Build proper cross-reference
                    if ref['section_id']:
                        refuri = app.builder.get_relative_uri(docname, ref['docname']) + '#' + ref['section_id']
                    else:
                        refuri = app.builder.get_relative_uri(docname, ref['docname'])
                    
                    if ref['page_title'] != ref_text:
                        para += nodes.Text(ref['page_title'] + ' » ')

                    ref_node = nodes.reference(
                        '',
                        ref_text,
                        internal=True,
                        refuri=refuri,
                    )
                    para += ref_node
                    
                    seealso += para
                
                # Append to the existing desc_content
                desc_content += seealso


def purge_related_items(app, env, docname):
    """Remove related items when a document is removed."""
    if not hasattr(env, 'related_items'):
        return
    
    # Remove entries that reference the removed document
    for target, refs in list(env.related_items.items()):
        env.related_items[target] = [
            ref for ref in refs if ref['docname'] != docname
        ]
        if not env.related_items[target]:
            del env.related_items[target]


def merge_related_items(app, env, docnames, other):
    """Merge related items from parallel builds."""
    if not hasattr(other, 'related_items'):
        return
    
    if not hasattr(env, 'related_items'):
        env.related_items = {}
    
    for target, refs in other.related_items.items():
        if target not in env.related_items:
            env.related_items[target] = []
        env.related_items[target].extend(refs)


def setup(app):
    """Setup the extension."""
    app.add_directive('related', RelatedDirective)
    app.connect('doctree-resolved', process_related_items)
    app.connect('env-purge-doc', purge_related_items)
    app.connect('env-merge-info', merge_related_items)
    
    return {
        'version': '0.1',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



"""DEPRECATED MODULE

Warning: This module is deprecated and will be removed in a future version.

It isn't loaded by __init__ -- it's intended to handle cases where users:
import pyterrier.batchretrieve
or
from pyterrier.batchretrieve import X
"""
from deprecated import deprecated
import pyterrier as pt

BatchRetrieve = pt.terrier.BatchRetrieve # already has @deprecated
FeaturesBatchRetrieve = pt.terrier.FeaturesBatchRetrieve # already has @deprecated
BatchRetrieveBase = pt.terrier.BatchRetrieveBase # already has @deprecated

@deprecated(version='0.11.0', reason="use pt.terrier.TextScorer() instead")
class TextScorer(pt.terrier.TextScorer):
    @staticmethod
    @deprecated(version='0.11.0', reason="use pt.terrier.TextScorer.from_dataset() instead")
    def from_dataset(*args, **kwargs):
        return pt.terrier.TextScorer.from_dataset(*args, **kwargs)


_from_dataset = deprecated(version='0.11.0', reason="use pt.datasets.transformer_from_dataset() instead")(pt.datasets.transformer_from_dataset)



"""Tools for working with the PyTerrier documentation."""
import zlib
from datetime import datetime
from pathlib import Path
from typing import Optional, Union, Dict, Tuple

import pyterrier as pt
import requests

_BASE_URL = 'https://pyterrier.readthedocs.io/en/latest/'

_cached_objects_inv = None


def _is_old(path: Path, url: str) -> bool:
    file_last_updated = path.stat().st_mtime
    if file_last_updated < datetime.now().timestamp() - 86400:
        # check at most once per day
        return False
    try:
        response = requests.head(url)
        response.raise_for_status()
        last_modified = response.headers.get('Last-Modified')
        if last_modified:
            last_modified_date = datetime.strptime(last_modified, '%a, %d %b %Y %H:%M:%S GMT')
            is_old = file_last_updated < last_modified_date.timestamp()
            if not is_old:
                path.touch() # reset the st_mtime
            return is_old
    except requests.RequestException:
        pass  # If we can't fetch the URL, assume it's old
    return True


def objects_inv() -> dict:
    """Returns a dictionary mapping objects IDs in the PyTerrier documentation to their documentation pages."""
    global _cached_objects_inv
    if _cached_objects_inv is not None:
        return _cached_objects_inv
    objects_inv_path = Path(pt.io.pyterrier_home()) / 'documentation' / 'objects.inv'
    objects_inv_url = _BASE_URL + 'objects.inv'
    if not objects_inv_path.exists() or _is_old(objects_inv_path, objects_inv_url):
        if not objects_inv_path.parent.exists():
            objects_inv_path.parent.mkdir(parents=True, exist_ok=True)
        pt.io.download(objects_inv_url, str(objects_inv_path), verbose=False, headers={"User-Agent": "curl/7.81.0"})
    # parse the objects.inv file
    with objects_inv_path.open('rb') as f:
        for _ in range(4):
            lineb = f.readline()
            if not lineb.startswith(b'#'):
                raise ValueError(f'Invalid objects.inv file: expected comment line, got {lineb!r}')
        data = f.read()
    lines = zlib.decompress(data).decode('utf-8').splitlines()
    objects : Dict[Tuple[str,str], Tuple[str, str, int]] = {}
    for line in lines:
        if line.startswith('#'):
            continue  # skip comments
        parts = line.strip().split(maxsplit=4)
        if len(parts) < 5:
            continue  # skip malformed lines
        name, type_, priority, url, title = parts
        ipriority = int(priority)
        if (type_, name) not in objects or objects[(type_, name)][2] > ipriority:
            objects[type_, name] = (_BASE_URL + url, title, ipriority)
    _cached_objects_inv = objects
    return _cached_objects_inv


def url_for_class(cls: Union[type, object]) -> Optional[str]:
    """Returns the URL of the documentation page for the specified class."""
    objects = objects_inv()
    if not isinstance(cls, type):
        cls = cls.__class__
    cls_name = f'{cls.__module__}.{cls.__name__}'
    key = ('py:class', cls_name)
    if key in objects:
        return objects[key][0]
    return None



__version__ = '1.0.0'
# NB: version number must be the first line and must use single quotes for the sed expression in .github/workflows/publish-to-pypi.yml

from typing import Any
from deprecated import deprecated

from pyterrier import model, utils, validate, testing
from pyterrier.transformer import Transformer, Estimator, Indexer
from pyterrier._ops import RankCutoff, Compose
from pyterrier._artifact import Artifact

from pyterrier import java

from pyterrier import terrier
from pyterrier.terrier import BatchRetrieve, TerrierRetrieve, FeaturesBatchRetrieve, IndexFactory, run, rewrite, index, FilesIndexer, TRECCollectionIndexer, DFIndexer, DFIndexUtils, IterDictIndexer, IndexingType, TerrierStemmer, TerrierStopwords, TerrierTokeniser
from pyterrier._evaluation import Experiment, GridScan, GridSearch, Evaluate, KFoldGridSearch

from pyterrier import debug
from pyterrier import documentation
from pyterrier import io
from pyterrier import inspect
from pyterrier import measures
from pyterrier import new
from pyterrier import ltr
from pyterrier import pipelines
from pyterrier import text
from pyterrier import transformer
from pyterrier import schematic
from pyterrier import datasets
from pyterrier.datasets import get_dataset, find_datasets, list_datasets
from pyterrier import apply as _apply_base

# old name
Utils = utils

# will be set in terrier.terrier.java once java is loaded
IndexRef = None

# these will be set once _() runs, but we need to define them here to get type checking to work properly
tqdm: Any
apply: _apply_base._apply


# deprecated functions explored to the main namespace, which will be removed in a future version
init = java.legacy_init # java.legacy_init raises a deprecated warning internally
started = deprecated(version='0.11.0', reason="use pt.java.started() instead")(java.started)
logging = deprecated(version='0.11.0', reason="use pt.java.set_log_level(...) instead")(java.set_log_level)
version = deprecated(version='0.11.0', reason="use pt.terrier.version() instead")(terrier.version)
check_version = deprecated(version='0.11.0', reason="use pt.terrier.check_version(...) instead")(terrier.check_version)
extend_classpath = deprecated(version='0.11.0', reason="use pt.terrier.extend_classpath(...) instead")(terrier.extend_classpath)
set_tqdm = deprecated(version='0.11.0', reason="use pt.utils.set_tqdm(...) instead")(utils.set_tqdm)
set_property = deprecated(version='0.11.0', reason="use pt.terrier.set_property(...) instead")(terrier.set_property)
set_properties = deprecated(version='0.11.0', reason="use pt.terrier.set_properties(...) instead")(terrier.set_properties)
redirect_stdouterr = deprecated(version='0.11.0', reason="use pt.java.redirect_stdouterr(...) instead")(java.redirect_stdouterr)
autoclass = deprecated(version='0.11.0', reason="use pt.java.autoclass(...) instead")(java.autoclass)
cast = deprecated(version='0.11.0', reason="use pt.java.cast(...) instead")(java.cast)


__all__ = [
    'java', 'terrier', 'debug', 'documentation', 'io', 'inspect', 'measures', 'model', 'new', 'ltr', 'pipelines', 'schematic',
    'text', 'transformer', 'datasets', 'validate', 'testing', 'get_dataset', 'find_datasets', 'list_datasets', 'Experiment', 'GridScan',
    'GridSearch', 'KFoldGridSearch', 'Evaluate',
    'utils', 'Utils', 'Transformer', 'Estimator', 'Indexer', 'Artifact',
    'RankCutoff', 'Compose',
    'BatchRetrieve', 'TerrierRetrieve', 'FeaturesBatchRetrieve', 'IndexFactory',
    'run', 'rewrite', 'index', 'FilesIndexer', 'TRECCollectionIndexer', 'DFIndexer', 'DFIndexUtils', 'IterDictIndexer',
    'IndexingType', 'TerrierStemmer', 'TerrierStopwords', 'TerrierTokeniser',
    'IndexRef', 'ApplicationSetup', 'properties',

    # Deprecated:
    'init', 'started', 'logging', 'version', 'check_version', 'extend_classpath', 'set_tqdm', 'set_property', 'set_properties',
    'redirect_stdouterr', 'autoclass', 'cast',

    # Entry point modules (appended loaded below):
]


# Additional setup performed in a function to avoid polluting the namespace with other imports like platform
def _():
    from warnings import warn
    import platform
    from packaging.version import Version

    # check python version
    if Version(platform.python_version()) < Version('3.7.0'):
        raise RuntimeError("From PyTerrier 0.8, Python 3.7 minimum is required, you currently have %s" % platform.python_version())

    globs = globals()

    # Load the _apply object as pt.apply so that the dynamic __getattr__ methods work
    from pyterrier.apply import _apply
    globs['apply'] = _apply()
    __all__.append('apply')

    # load modules defined as package entry points into the global pyterrier namespace
    for entry_point in utils.entry_points('pyterrier.modules'):
        if entry_point.name in globs:
            warn(f'skipping loading {entry_point} because a module with this name is already loaded.')
            continue
        module = entry_point.load()
        if callable(module): # if the entry point refers to an function/class, call it to get the module
            module = module()
        globs[entry_point.name] = module
        __all__.append(entry_point.name)

    # guess the environment and set an appropriate tqdm as pt.tqdm
    utils.set_tqdm()
_()



from typing import List, Union, Literal

import pandas as pd
import pyterrier as pt
from pyterrier.terrier._index import TerrierIndex


class TerrierTextLoader(pt.Transformer):
    """A transformer that loads textual metadata from a Terrier index into a DataFrame by docid or docno."""
    schematic = {'label': 'TextLoader'}
    def __init__(
        self,
        index: TerrierIndex,
        fields: Union[List[str], str, Literal['*']] = '*',
        *,
        verbose=False
    ):
        """Initialise the transformer with the index to load metadata from.

        Args:
            index (pyterrier.terrier.J.Index): The index to load metadata from.
            fields: The fields to load from the index. If '*', all fields will be loaded.
            verbose: Whether to print debug information.
        """
        metaindex = index.index_obj().getMetaIndex()

        if metaindex is None:
            raise ValueError(f"Index {index} does not have a metaindex")

        available_fields = list(metaindex.getKeys())
        if fields == '*':
            fields = available_fields
        else:
            if isinstance(fields, str):
                fields = [fields]
            missing_fields = set(fields) - set(available_fields)
            if missing_fields:
                raise ValueError(f"Index from {index} did not have requested metaindex keys {list(missing_fields)}. "
                                 f"Keys present in metaindex are {available_fields}")
        self._index = index
        self.fields = fields
        self.verbose = verbose

    def transform(self, inp: pd.DataFrame) -> pd.DataFrame:
        """Load metadata from the index into the input DataFrame.

        Args:
            inp: The input DataFrame. Must contain either 'docid' or 'docno'.

        Returns:
            A new DataFrame with the metadata columns appended.
        """
        if 'docno' not in inp.columns and 'docid' not in inp.columns:
            raise ValueError(f"Neither docid nor docno are in the input dataframe, found {list(inp.columns)}")

        metaindex = self._index.index_obj().getMetaIndex()

        # Get the docids
        if "docid" not in inp.columns:
            # Look up docids by docno
            docids = inp.docno.map(lambda docno: metaindex.getDocument("docno", docno))
        else:
            # Use the provided docids
            docids = inp.docid

        # Look up the metadata and build a new frame to append
        docids = docids.values.tolist() # getItems expects a list
        metadata_matrix = metaindex.getItems(self.fields, docids) # indexed by docid then keys
        metadata_frame = pd.DataFrame(metadata_matrix, columns=self.fields)

        # append the input and metadata frames
        inp = inp.drop(columns=self.fields, errors='ignore') # make sure we don't end up with duplicates
        inp = inp.reset_index(drop=True) # reset the index to default (matching metadata_frame)
        return pd.concat([inp, metadata_frame], axis='columns')

    def fuse_rank_cutoff(self, k):
        return pt.RankCutoff(k) >> self

def terrier_text_loader(
    index,
    fields: Union[List[str], str, Literal['*']] = '*',
    *,
    verbose=False
) -> TerrierTextLoader:
    """Create a transformer that loads textual metadata from a Terrier index into a DataFrame by docid or docno.

    Args:
        index (str or pyterrier.terrier.J.IndexRef or pyterrier.terrier.J.Index): The index to load metadata from.
        fields: The fields to load from the index. If '*', all fields will be loaded.
        verbose: Whether to print debug information.
    """
    index = pt.terrier.TerrierIndex.coerce(index)
    return TerrierTextLoader(index, fields, verbose=verbose)



""" This module contains utilities for testing PyTerrier components and functionalities."""
import os
import unittest
import warnings

import pyterrier as pt

class TransformerTestCase(unittest.TestCase):
    """A base test case for testing PyTerrier transformers."""

    @classmethod
    def setUpClass(cls):
        cls.transformer = cls.get_transformer()
        cls.io_configs = []

    @classmethod
    def tearDownClass(cls):
        if os.environ.get('PT_TEST_TRANSFORMER_REPORTS', '1') == '1':
            if cls.io_configs:
                io_configs = "\n".join([f"    - {' '.join(inp)} -> {' '.join(out)}" for inp, out in cls.io_configs])
            else:
                io_configs = '    [Unavailable (see test failures)]'
            warnings.warn('You can hide TransfomerTestCase reports by setting PT_TEST_TRANSFORMER_REPORTS=0')
            warnings.warn(f'''INFO: TransfomerTestCase Report for {str(cls.transformer)}
  Input/Output Configurations:
{io_configs}
            ''')
        cls.transformer = None

    @staticmethod
    def get_transformer():
        """Return the transformer instance to be tested."""
        raise NotImplementedError()

    def test_inspect_io(self):
        inputs = []
        with self.subTest("pt.inspect.transformer_inputs"):
            inputs = pt.inspect.transformer_inputs(self.transformer)
            self.assertIsInstance(inputs, list, "transformer_inputs not a list")
            if len(inputs) > 0:
                self.assertTrue(all(isinstance(i, list) for i in inputs), "transformer_inputs not a list of lists")
                self.assertTrue(all(isinstance(x, str) for i in inputs for x in i), "transformer_inputs not a list of list of strings")
            else:
                warnings.warn('Transformer does not support any input configurations. Are you sure this is right?')

        for input_config in inputs:
            with self.subTest(**{'pt.inspect.transformer_outputs': {'input_columns': input_config}}):
                outputs = pt.inspect.transformer_outputs(self.transformer, input_columns=input_config)
                self.assertIsInstance(outputs, list, "transformer_outputs not a list")
                self.assertTrue(all(isinstance(o, str) for o in outputs), "transformer_outputs not a list of strings")
                self.io_configs.append((input_config, outputs))

    # Disable for now:
    # def test_inspect_attr(self):
    #     with self.subTest('pt.inspect.transformer_attributes'):
    #         attrs = pt.inspect.transformer_attributes(self.transformer)
    #         self.assertIsInstance(attrs, list, "transformer_attributes not a list")
    #         self.assertTrue(all(isinstance(a, pt.inspect.TransformerAttribute) for a in attrs), "transformer_attributes not a list of TransformerAttribute")
    #         self.attrs.extend([a.name for a in attrs])

    #     with self.subTest('pt.inspect.transformer_apply_attributes'):
    #         new_transformer = pt.inspect.transformer_apply_attributes(self.transformer)
    #         self.assertFalse(self.transformer is new_transformer, "transformer_apply_attributes returned the same instance")

    #     with self.subTest('pt.inspect.subtransformers'):
    #         subtransformers = pt.inspect.subtransformers(self.transformer)
    #         self.assertIsInstance(subtransformers, dict, "subtransformers not a dict")

    def test_schematic(self):
        schematic = None
        with self.subTest('pt.schematic.transformer_schematic'):
            schematic = pt.schematic.transformer_schematic(self.transformer)
            if schematic['type'] == 'transformer' and len(schematic['label']) > 20:
                self.fail(f"Schematic label ({schematic['label']}) is long ({len(schematic['label'])} characters). This may affect rendering. Try to keep it below 20 characters.")

        if schematic is not None:
            with self.subTest('pt.schematic.draw'):
                schematic = pt.schematic.draw(schematic)


def transformer_test_class(fn):
    class _TransformerTestCase(TransformerTestCase):
        @staticmethod
        def get_transformer():
            return fn()
    return _TransformerTestCase



def _save_state(param_dict):
    rtr = []
    for tran, param_set in param_dict.items():
        for param_name in param_set:
            rtr.append((tran, param_name, tran.get_parameter(param_name)))
    return rtr


def _restore_state(param_state):
    for (tran, param_name, param_value) in param_state:
        tran.set_parameter(param_name, param_value)


import pyterrier as pt
import unittest
from .base import BaseTestCase
import os
import pandas as pd
from pytest import warns

class TestTransformer(BaseTestCase):

    def test_call(self):
        inputDocs = pt.new.ranked_documents([[2, 1], [2]], qid=["q100", "q10"])
        t = pt.Transformer.from_df(inputDocs)
        self.assertEqual(2, len(t(pt.new.queries(['a'], qid=['q100']))))
        self.assertEqual(1, len(t(pt.new.queries(['a'], qid=['q10']))))
        self.assertEqual(2, len(t([{'qid' : 'q100'}])))
        self.assertEqual(1, len(t([{'qid' : 'q10'}])))

    def test_is_transformer(self):
        class IncompleteTransformer(pt.Transformer):
            pass # doesnt implement .transform() or .transform_iter()
        class MyTransformer1(pt.Transformer):
            def transform(self, df):
                pass
        class MyTransformer1a(pt.Transformer):
            def transform_iter(self, df):
                pass
        class MyTransformer3(pt.Indexer):
            pass # indexers dont need a transform
        class MyTransformer4(pt.Estimator):
            def transform(self, df):
                pass

        with self.assertRaises(NotImplementedError):
            IncompleteTransformer()

        # check normal API
        for T in [MyTransformer1, MyTransformer1a, MyTransformer3, MyTransformer4]:
            self.assertTrue(pt.transformer.is_transformer(T()))

        # check for inspectability
        iter_notinspectable = MyTransformer1a()
        self.assertTrue(hasattr(iter_notinspectable, 'transform_outputs'))
        with self.assertRaises(pt.inspect.InspectError):
            iter_notinspectable.transform_outputs(['a'])

        self.assertFalse(hasattr(MyTransformer1(), 'transform_outputs'))
